# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
# pylint: disable=redefined-outer-name,import-error,protected-access
# flake8: noqa: B009

"""Tests for the aea_ledger_ethereum_flashbots package."""

from typing import Tuple
from unittest.mock import ANY, MagicMock, patch

import pytest
from aea_ledger_ethereum_flashbots.ethereum_flashbots import EthereumFlashbotApi
from eth_account import Account
from flashbots.types import FlashbotsBundleRawTx
from hexbytes import HexBytes
from web3.exceptions import TransactionNotFound
from web3.types import TxReceipt

_DUMMY_FLASHBOTS_BUILDERS = [
    ["dummy", "dummy_uri1"],
    ["builder0x69", "dummy_uri2"],
]


@pytest.fixture
def ethereum_flashbot_api() -> EthereumFlashbotApi:
    """Get the ethereum flashbot API."""

    return EthereumFlashbotApi(flashbots_builders=_DUMMY_FLASHBOTS_BUILDERS)


def test_init_with_signature_private_key() -> None:
    """Test init with signature private key."""
    signature_private_key = "my private key"
    with patch.object(
        Account, "from_key", side_effect=lambda private_key: MagicMock()
    ) as account_from_key_mock:
        EthereumFlashbotApi(
            signature_private_key=signature_private_key,
            flashbots_builders=_DUMMY_FLASHBOTS_BUILDERS,
        )
        assert account_from_key_mock.called_once_with(signature_private_key)


def test_init_without_signature_private_key() -> None:
    """Test init without signature private key."""
    with patch.object(
        Account, "create", side_effect=MagicMock()
    ) as account_create_mock:
        EthereumFlashbotApi(flashbots_builders=_DUMMY_FLASHBOTS_BUILDERS)
        assert account_create_mock.called_once_with()


@pytest.mark.parametrize("signed_txs", (("0x1234", "0x0000", "0x5232"), ("0x1234",)))
def test_bundle_transactions(ethereum_flashbot_api, signed_txs: Tuple[str]) -> None:
    """Test bundle transactions."""
    dummy_signed_transactions = [
        dict(raw_transaction=signed_tx) for signed_tx in signed_txs
    ]
    bundle = ethereum_flashbot_api.bundle_transactions(dummy_signed_transactions)
    actual_bundle = [tx.get("signed_transaction", None) for tx in bundle]
    expected_bundle = [
        HexBytes(tx.get("raw_transaction", None)) for tx in dummy_signed_transactions
    ]
    assert expected_bundle == actual_bundle


def test_simulate_with_successful_simulation(
    ethereum_flashbot_api,
) -> None:
    """Test simulate with successful simulation."""
    # mock
    response_mock = MagicMock()
    default_builder = _DUMMY_FLASHBOTS_BUILDERS[0][0]
    flashbots_module = getattr(
        ethereum_flashbot_api._builder_to_web3[default_builder], "flashbots"
    )
    flashbots_module.simulate = MagicMock(return_value=response_mock)

    # run
    bundle = [FlashbotsBundleRawTx(signed_transaction=HexBytes("0x1234"))]
    success = ethereum_flashbot_api.simulate(bundle, 123)

    # check
    flashbots_module.simulate.assert_called_once_with(bundle, 123)
    assert success


def test_simulate_with_failed_simulation(ethereum_flashbot_api) -> None:
    """Test simulate with failed simulation."""
    # mock
    response_mock = MagicMock(side_effect=Exception)
    ethereum_flashbot_api.flashbots.simulate = MagicMock(side_effect=response_mock)

    # run
    bundle = FlashbotsBundleRawTx[{"signed_transaction": "0x1234"}]
    success = ethereum_flashbot_api.simulate(bundle, 123)

    # check
    ethereum_flashbot_api.flashbots.simulate.assert_called_once_with(bundle, 123)
    assert not success


def test_send_bundle_with_successful_transaction(ethereum_flashbot_api) -> None:
    """Test send bundle with successful transaction."""
    current_block = 122
    send_bundle = [FlashbotsBundleRawTx(signed_transaction=HexBytes("0x1234"))]
    res_bundle = [{"hash": b"0x1234"}]
    response_mock = MagicMock()
    response_mock.wait = MagicMock()
    response_mock.receipts = MagicMock(return_value=[TxReceipt(blockNumber=1)])
    response_mock.bundle = res_bundle
    ethereum_flashbot_api.api.eth.get_block_number = MagicMock(
        return_value=current_block
    )
    ethereum_flashbot_api.flashbots.simulate = MagicMock(return_value=True)
    ethereum_flashbot_api.flashbots.send_bundle = MagicMock(return_value=response_mock)

    # run
    target_blocks = [123]
    tx_hashes = ethereum_flashbot_api.send_bundle(send_bundle, target_blocks)

    # check
    ethereum_flashbot_api.flashbots.simulate.assert_called_once_with(
        send_bundle, current_block
    )
    ethereum_flashbot_api.flashbots.send_bundle.assert_called_once_with(
        send_bundle, target_blocks[0], opts={"replacementUuid": ANY}
    )
    assert response_mock.wait.called
    assert tx_hashes == [tx["hash"].hex() for tx in response_mock.bundle]


def test_bundle_transactions_with_empty_list(ethereum_flashbot_api) -> None:
    """Test bundle transactions with an empty list of signed transactions."""
    signed_transactions = []
    bundle = ethereum_flashbot_api.bundle_transactions(signed_transactions)
    assert len(bundle) == 0


def test_simulate_with_empty_bundle(ethereum_flashbot_api) -> None:
    """Test simulate with an empty bundle."""
    bundle = []
    success = ethereum_flashbot_api.simulate(bundle, 123)
    assert not success


def test_simulate_with_invalid_target_blocks(ethereum_flashbot_api) -> None:
    """Test simulate with invalid target blocks."""
    bundle = [FlashbotsBundleRawTx(signed_transaction=HexBytes("0x1234"))]
    success = ethereum_flashbot_api.simulate(bundle, -1)
    assert not success


def test_bundle_and_send_with_successful_transaction(ethereum_flashbot_api) -> None:
    """Test bundle and send with successful transaction."""
    # mock
    response_mock = MagicMock()
    response_mock.wait = MagicMock()
    response_mock.receipts = MagicMock(
        return_value=[TxReceipt(blockNumber=1), TxReceipt(blockNumber=2)]
    )
    response_mock.bundle = [{"hash": b"0x1234"}, {"hash": b"0x5678"}]
    ethereum_flashbot_api._get_next_blocks = MagicMock(return_value=1)
    ethereum_flashbot_api.api.eth.get_block_number = MagicMock(return_value=1)
    ethereum_flashbot_api.flashbots.simulate = MagicMock(return_value=True)
    ethereum_flashbot_api.flashbots.send_bundle = MagicMock(return_value=response_mock)

    # run
    signed_transactions = [
        dict(raw_transaction="0x1234"),
        dict(raw_transaction="0x5678"),
    ]
    target_blocks = [123]

    # check
    tx_hashes = ethereum_flashbot_api.send_signed_transactions(
        signed_transactions,
        target_blocks=target_blocks,
    )
    ethereum_flashbot_api.flashbots.simulate.assert_called_once()
    ethereum_flashbot_api.flashbots.send_bundle.assert_called_once()
    assert response_mock.wait.called
    assert tx_hashes == [tx["hash"].hex() for tx in response_mock.bundle]


def test_bundle_and_send_with_failed_simulation(ethereum_flashbot_api) -> None:
    """Test bundle and send with failed simulation."""
    # mock
    response_mock = MagicMock()
    response_mock.wait = MagicMock()
    response_mock.bundle_hash = MagicMock()
    response_mock.receipts = MagicMock(
        side_effect=TransactionNotFound("Receipt not found")
    )
    ethereum_flashbot_api._get_next_blocks = MagicMock(return_value=1)
    ethereum_flashbot_api.api.eth.get_block_number = MagicMock(return_value=1)
    ethereum_flashbot_api.flashbots.simulate = MagicMock(return_value=True)
    ethereum_flashbot_api.flashbots.send_bundle = MagicMock(return_value=response_mock)
    ethereum_flashbot_api.flashbots.cancel_bundles = MagicMock()
    ethereum_flashbot_api.api.toHex = MagicMock()
    ethereum_flashbot_api.flashbots.get_bundle_stats_v2 = MagicMock(
        return_value=response_mock
    )

    # run
    signed_transactions = [
        dict(raw_transaction="0x1234"),
        dict(raw_transaction="0x5678"),
    ]
    target_blocks = [123]
    tx_hashes = ethereum_flashbot_api.send_signed_transactions(
        signed_transactions,
        target_blocks=target_blocks,
        raise_on_try=True,
    )

    # check
    ethereum_flashbot_api.flashbots.simulate.assert_called_once()
    ethereum_flashbot_api.flashbots.send_bundle.assert_called_once()
    ethereum_flashbot_api.flashbots.cancel_bundles.assert_called_once()
    assert tx_hashes is None


def test_send_bundle_with_failed_simulation(ethereum_flashbot_api) -> None:
    """Test send bundle with failed simulation."""
    # mock
    response_mock = MagicMock()
    response_mock.wait = MagicMock()
    response_mock.bundle_hash = MagicMock()
    response_mock.receipts = MagicMock(side_effect=TransactionNotFound)
    ethereum_flashbot_api.api.eth.get_block_number = MagicMock(return_value=1)
    ethereum_flashbot_api.simulate = MagicMock(return_value=False)
    target_blocks = [0]

    # run
    tx_hashes = ethereum_flashbot_api.send_bundle(MagicMock(), target_blocks)

    # check
    assert tx_hashes is None


def test_send_bundle_with_failed_simulation_and_raise(ethereum_flashbot_api) -> None:
    """Test send bundle with failed simulation and should raise."""
    # mock
    response_mock = MagicMock()
    response_mock.wait = MagicMock()
    response_mock.bundle_hash = MagicMock()
    response_mock.receipts = MagicMock(side_effect=TransactionNotFound)
    ethereum_flashbot_api.api.eth.get_block_number = MagicMock(return_value=0)
    ethereum_flashbot_api.simulate = MagicMock(return_value=False)
    raise_on_failed_simulation = True
    target_blocks = [1]

    # run
    with pytest.raises(ValueError):
        ethereum_flashbot_api.send_bundle(
            MagicMock(), target_blocks, raise_on_failed_simulation
        )


def test_send_bundle_with_bad_target_block(ethereum_flashbot_api) -> None:
    """Test send bundle with an old target block should return None."""
    # mock
    response_mock = MagicMock()
    response_mock.wait = MagicMock()
    response_mock.bundle_hash = MagicMock()
    response_mock.receipts = MagicMock(side_effect=TransactionNotFound)
    ethereum_flashbot_api.api.eth.get_block_number = MagicMock(return_value=1)
    ethereum_flashbot_api.simulate = MagicMock(return_value=True)
    raise_on_failed_simulation = True
    target_blocks = [0]

    # run
    res = ethereum_flashbot_api.send_bundle(
        MagicMock(), target_blocks, raise_on_failed_simulation
    )
    assert res is None, "Should return None if target block is old"



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains types and helpers for ACN Proof-of-Representation."""

from typing import Optional

from aea.common import PathLike
from aea.crypto.registries import make_ledger_api
from aea.helpers.base import CertRequest, SimpleId, SimpleIdOrStr


class AgentRecord:
    """Agent Proof-of-Representation to representative."""

    __slots__ = (
        "_address",
        "_representative_public_key",
        "_identifier",
        "_ledger_id",
        "_not_before",
        "_not_after",
        "_message_format",
        "_signature",
        "_message",
        "_public_key",
    )

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        address: str,
        representative_public_key: str,
        identifier: SimpleIdOrStr,
        ledger_id: SimpleIdOrStr,
        not_before: str,
        not_after: str,
        message_format: str,
        signature: str,
    ) -> None:
        """
        Initialize the AgentRecord

        :param address: agent address
        :param representative_public_key: representative's public key
        :param identifier: certificate identifier.
        :param ledger_id: ledger identifier the request is referring to.
        :param not_before: specify the lower bound for certificate validity. If it is a string, it must follow the format: 'YYYY-MM-DD'. It will be interpreted as timezone UTC-0.
        :param not_after: specify the lower bound for certificate validity. If it is a string, it must follow the format: 'YYYY-MM-DD'. It will be interpreted as timezone UTC-0.
        :param message_format: message format used for signing
        :param signature: proof-of-representation of this AgentRecord
        """
        self._address = address
        self._representative_public_key = representative_public_key
        self._identifier = str(SimpleId(identifier))
        self._ledger_id = str(SimpleId(ledger_id))
        self._not_before = not_before
        self._not_after = not_after
        self._message_format = message_format
        self._signature = signature
        self._message = CertRequest.construct_message(
            self.representative_public_key,
            self.identifier,
            self.not_before,
            self.not_after,
            self.message_format,
        )
        self._public_key = self._check_validity()

    def _check_validity(self) -> str:
        """
        Checks validity of record.

        Specifically:
        - if ledger_id is valid
        - if agent signed the message

        :return: agent public key
        """
        ledger_api = make_ledger_api(self.ledger_id)
        public_keys = ledger_api.recover_public_keys_from_message(
            self.message, self.signature
        )
        if len(public_keys) == 0:
            raise ValueError("Malformed signature!")  # pragma: no cover
        public_key: Optional[str] = None
        for public_key_ in public_keys:
            address = ledger_api.get_address_from_public_key(public_key_)
            if address == self.address:
                public_key = public_key_
                break
        if public_key is None:
            raise ValueError(
                "Invalid signature for provided representative_public_key and agent address!"
            )
        return public_key

    @property
    def address(self) -> str:
        """Get agent address"""
        return self._address

    @property
    def public_key(self) -> str:
        """Get agent public key"""
        return self._public_key

    @property
    def representative_public_key(self) -> str:
        """Get agent representative's public key"""
        return self._representative_public_key

    @property
    def signature(self) -> str:
        """Get record signature"""
        return self._signature

    @property
    def message(self) -> bytes:
        """Get the message."""
        return self._message

    @property
    def identifier(self) -> SimpleIdOrStr:
        """Get the identifier."""
        return self._identifier

    @property
    def ledger_id(self) -> SimpleIdOrStr:
        """Get ledger id."""
        return self._ledger_id

    @property
    def not_before(self) -> str:
        """Get the not_before field."""
        return self._not_before

    @property
    def not_after(self) -> str:
        """Get the not_after field."""
        return self._not_after

    @property
    def message_format(self) -> str:
        """Get the message format."""
        return self._message_format

    def __str__(self) -> str:  # pragma: no cover
        """Get string representation."""
        return f"(address={self.address}, public_key={self.public_key}, representative_public_key={self.representative_public_key}, signature={self.signature}, ledger_id={self.ledger_id})"

    @classmethod
    def from_cert_request(
        cls,
        cert_request: CertRequest,
        address: str,
        representative_public_key: str,
        data_dir: Optional[PathLike] = None,
    ) -> "AgentRecord":
        """Get agent record from cert request."""
        signature = cert_request.get_signature(data_dir)
        record = cls(
            address,
            representative_public_key,
            cert_request.identifier,
            cert_request.ledger_id,
            cert_request.not_before_string,
            cert_request.not_after_string,
            cert_request.message_format,
            signature,
        )
        return record



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test `aea hash` command."""

from typing import Set

import click
import pytest

from aea.configurations.data_types import PackageId
from aea.helpers.dependency_tree import DependencyTree
from aea.test_tools.exceptions import AEATestingException
from aea.test_tools.test_cases import BaseAEATestCase


class TestHashAll(BaseAEATestCase):
    """Test `aea hash all`"""

    use_packages_dir: bool = True

    def test_invocation(
        self,
    ) -> None:
        """Test command invocation."""

        tree = DependencyTree.generate(self.packages_dir_path)
        packages: Set[PackageId] = set()
        for level in tree:
            packages.update(level)

        result = self.run_cli_command(
            "hash", "all", "--packages-dir", str(self.packages_dir_path)
        )

        assert result.exit_code == 0, result.output
        assert all(
            f"Processing package {package_id.name} of type {package_id.package_type}"
            in result.output
            for package_id in packages
        )

    def test_empty_packages_dir(
        self,
    ) -> None:
        """Test packages directory being empty"""

        with pytest.raises(
            AEATestingException, match="ValueError: Could not find any package in"
        ):
            self.run_cli_command("hash", "all", "--packages-dir", str(self.t))

    def test_vendor_flag(
        self,
    ) -> None:
        """Test `--vendor` flag"""

        vendor = "valory"
        tree = DependencyTree.generate(self.packages_dir_path)
        packages: Set[PackageId] = set()
        for level in tree:
            packages.update(level)

        result = self.run_cli_command(
            "hash",
            "all",
            "--packages-dir",
            str(self.packages_dir_path),
            "--vendor",
            vendor,
        )

        assert result.exit_code == 0, result.output
        assert all(
            f"Skipping the hash update for {package_id.name} of type {package_id.package_type}"
            in result.output
            for package_id in packages
            if package_id.author != vendor
        ), result.output


class TestHashOne(BaseAEATestCase):
    """Test `aea hash one`"""

    use_packages_dir: bool = True

    def test_run(
        self,
    ) -> None:
        """Test command invocation."""

        temp_file = self.t / "file.txt"
        temp_file.write_text("Hello, World!")

        result = self.run_cli_command("hash", "one", str(temp_file))

        assert result.exit_code == 0, result.output
        assert (
            "Hash : bafybeibv2rf6ebovxm3mnr2bottqerc3p23uezk376pyaob4mu6qyddgw4"
            in result.output
        )

    def test_no_wrap(
        self,
    ) -> None:
        """Test command invocation with `--no-wrap` flag."""

        temp_file = self.t / "file.txt"
        temp_file.write_text("Hello, World!")

        result = self.run_cli_command("hash", "one", str(temp_file), "--no-wrap")

        assert result.exit_code == 0, result.output
        assert (
            "Hash : bafybeifkhqxtv22r566qadr6owj25yqdme2wecr4qsbehnh5wztzjb26le"
            in result.output
        )


class TestHashConversion(BaseAEATestCase):
    """Test hash conversion."""

    v0_hash = "QmZoBFCYa4gJRYhpKuVWXon461apqQYKb7FUqS2h9HPM3e"
    v1_hash = "bafybeifkhqxtv22r566qadr6owj25yqdme2wecr4qsbehnh5wztzjb26le"

    def test_to_v1(
        self,
    ) -> None:
        """Test `to-v1` command."""

        result = self.run_cli_command("hash", "to-v1", self.v0_hash)

        assert result.exit_code == 0, result.output
        assert self.v1_hash in result.output, result.output

    def test_to_v0(
        self,
    ) -> None:
        """Test `to-v0` command."""

        result = self.run_cli_command(
            "hash",
            "to-v0",
            self.v1_hash,
        )

        assert result.exit_code == 0, result.output
        assert self.v0_hash in result.output, result.output

    def test_to_v1_failure(
        self,
    ) -> None:
        """Test `to-v1` command."""

        with pytest.raises(click.ClickException, match=f"{self.v1_hash} is already v1"):
            self.run_cli_command("hash", "to-v1", self.v1_hash)

    def test_to_v0_failure(
        self,
    ) -> None:
        """Test `to-v1` command."""

        with pytest.raises(click.ClickException, match=f"{self.v0_hash} is already v0"):
            self.run_cli_command("hash", "to-v0", self.v0_hash)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the behaviour class of the echo skill."""

# pylint: skip-file

import inspect
import logging
import os
from pathlib import Path
from typing import cast
from unittest.mock import patch

from aea.test_tools.test_skill import BaseSkillTestCase

from packages.fetchai.skills.echo.behaviours import EchoBehaviour

CUR_PATH = os.path.dirname(inspect.getfile(inspect.currentframe()))  # type: ignore


class TestEchoBehaviour(BaseSkillTestCase):
    """Test EchoBehaviour behaviour of echo."""

    path_to_skill = Path(CUR_PATH, "..")
    is_agent_to_agent_messages = False

    @classmethod
    def setup_class(cls):
        """Setup the test class."""
        super().setup_class()
        cls.echo_behaviour = cast(
            EchoBehaviour, cls._skill.skill_context.behaviours.echo
        )
        cls.logger = cls._skill.skill_context.logger

    def test_setup(self):
        """Test the setup method of the echo behaviour."""
        # operation
        with patch.object(self.logger, "log") as mock_logger:
            assert self.echo_behaviour.setup() is None

        # after
        self.assert_quantity_in_outbox(0)

        mock_logger.assert_any_call(
            logging.INFO, "Echo Behaviour: setup method called."
        )

    def test_act(self):
        """Test the act method of the echo behaviour."""
        # operation
        with patch.object(self.logger, "log") as mock_logger:
            assert self.echo_behaviour.act() is None

        # after
        self.assert_quantity_in_outbox(0)

        mock_logger.assert_any_call(logging.INFO, "Echo Behaviour: act method called.")

    def test_teardown(self):
        """Test the teardown method of the echo behaviour."""
        # operation
        with patch.object(self.logger, "log") as mock_logger:
            assert self.echo_behaviour.teardown() is None

        # after
        self.assert_quantity_in_outbox(0)

        mock_logger.assert_any_call(
            logging.INFO, "Echo Behaviour: teardown method called."
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""
This script generates the IPFS hashes for all packages.

This script requires that you have IPFS installed:
- https://docs.ipfs.tech/guides/guides/install/
"""

import sys
import traceback
from pathlib import Path
from typing import Callable, Dict, Optional, Tuple, cast
from warnings import warn

import click

from aea.configurations.base import PackageConfiguration, PackageType
from aea.configurations.constants import (
    AGENT,
    PACKAGE_TYPE_TO_CONFIG_FILE,
    SCAFFOLD_PACKAGES,
    SCAFFOLD_PUBLIC_ID,
)
from aea.configurations.data_types import PackageId, PublicId
from aea.configurations.loader import load_configuration_object
from aea.helpers.cid import DEFAULT_ENCODING, to_v0, to_v1
from aea.helpers.dependency_tree import DependencyTree, dump_yaml, load_yaml, to_plural
from aea.helpers.fingerprint import update_fingerprint
from aea.helpers.ipfs.base import IPFSHashOnly
from aea.helpers.yaml_utils import yaml_dump, yaml_dump_all


def to_package_id(package_path: Path) -> Tuple[PackageId, Path]:
    """Extract the package type from the path."""
    item_type_plural = package_path.parent.name
    item_type_singular = item_type_plural[:-1]
    return (
        PackageId(
            PackageType(item_type_singular), PublicId.from_str(SCAFFOLD_PUBLIC_ID)
        ),
        package_path,
    )


def package_id_and_path(
    package_id: PackageId, packages_dir: Path
) -> Tuple[PackageId, Path]:
    """Extract the package type from the path."""
    package_path = (
        packages_dir
        / package_id.author
        / package_id.package_type.to_plural()
        / package_id.name
    )

    return package_id, package_path


def sort_configuration_file(config: PackageConfiguration) -> None:
    """Sort the order of the fields in the configuration files."""
    # load config file to get ignore patterns, dump again immediately to impose ordering
    if config.directory is None:
        raise ValueError("config.directory cannot be None.")

    configuration_filepath = config.directory / config.default_configuration_filename
    if config.package_type == PackageType.AGENT:
        json_data = config.ordered_json
        component_configurations = json_data.pop("component_configurations")
        yaml_dump_all(
            [json_data] + component_configurations, configuration_filepath.open("w")
        )
    elif config.package_type == PackageType.SERVICE:
        json_data = config.ordered_json
        overrides = json_data.pop("overrides")
        yaml_dump_all([json_data, *overrides], configuration_filepath.open("w"))
    else:
        yaml_dump(config.ordered_json, configuration_filepath.open("w"))


def hash_package(
    configuration: PackageConfiguration,
    package_type: PackageType,
    no_wrap: bool = False,
) -> Tuple[str, str]:
    """
    Hashes a package and its components.

    :param configuration: the package configuration.
    :param package_type: the package type.
    :param no_wrap: Whether to use the wrapper node or not.
    :return: the identifier of the hash (e.g. 'fetchai/protocols/default')
           | and the hash of the whole package.
    """
    # hash again to get outer hash (this time all files)
    # we still need to ignore some files
    #      use ignore patterns somehow
    # ignore_patterns = configuration.fingerprint_ignore_patterns # noqa: E800
    if configuration.directory is None:
        raise ValueError("configuration.directory cannot be None.")

    key = "/".join(
        [
            configuration.author,
            package_type.to_plural(),
            configuration.directory.name,
        ]
    )
    package_hash = IPFSHashOnly.hash_directory(
        str(configuration.directory), wrap=(not no_wrap)
    )

    return key, package_hash


def load_configuration(
    package_type: PackageType, package_path: Path
) -> PackageConfiguration:
    """
    Load a configuration, knowing the type and the path to the package root.

    :param package_type: the package type.
    :param package_path: the path to the package root.
    :return: the configuration object.
    """
    configuration_obj = load_configuration_object(package_type, package_path)
    configuration_obj._directory = package_path  # pylint: disable=protected-access
    return cast(PackageConfiguration, configuration_obj)


def update_public_id_hash(
    public_id_str: str,
    package_type: str,
    public_id_to_hash_mappings: Dict[PackageId, str],
) -> str:
    """Update hash for given public from available mappings."""
    public_id = PublicId.from_str(public_id_str).without_hash()
    package_id = PackageId(PackageType(package_type), public_id)
    return str(
        PublicId.from_json(
            {
                **public_id.json,
                "package_hash": cast(str, public_id_to_hash_mappings.get(package_id)),
            }
        )
    )


def extend_public_ids(
    item_config: Dict, public_id_to_hash_mappings: Dict[PackageId, str]
) -> None:
    """Extend public id with hashes for given item config."""
    for component_type in PACKAGE_TYPE_TO_CONFIG_FILE:
        if component_type == AGENT:
            continue

        components = to_plural(component_type)
        if components in item_config:
            item_config[components] = [
                update_public_id_hash(d, component_type, public_id_to_hash_mappings)
                for d in item_config.get(components, [])
            ]


def update_hashes(
    packages_dir: Path,
    no_wrap: bool = False,
    vendor: Optional[str] = None,
    config_loader: Callable[
        [PackageType, Path], PackageConfiguration
    ] = load_configuration,
) -> int:
    """Process all AEA packages and update fingerprint."""
    return_code = 0
    package_hashes: Dict[str, str] = {}

    try:
        public_id_to_hash_mappings: Dict = {}
        dependency_tree = DependencyTree.generate(packages_dir)
        if not dependency_tree:
            raise ValueError(f"Could not find any package in {packages_dir}")

        packages = [
            [package_id_and_path(package_id, packages_dir) for package_id in tree_level]
            for tree_level in dependency_tree
        ]
        packages[0] = packages[0] + list(map(to_package_id, SCAFFOLD_PACKAGES))
        for tree_level in packages:
            for package_id, package_path in tree_level:
                click.echo(
                    "Processing package {} of type {}".format(
                        package_path.name, package_id.package_type
                    )
                )

                config_file = package_path / cast(
                    str, PACKAGE_TYPE_TO_CONFIG_FILE.get(package_id.package_type.value)
                )
                item_config, extra_config = load_yaml(config_file)
                extend_public_ids(item_config, public_id_to_hash_mappings)
                dump_yaml(config_file, item_config, extra_config)

                configuration_obj = config_loader(package_id.package_type, package_path)
                sort_configuration_file(configuration_obj)
                update_fingerprint(configuration_obj)
                key, package_hash = hash_package(
                    configuration_obj, package_id.package_type, no_wrap=no_wrap
                )
                public_id_to_hash_mappings[package_id] = package_hash

                if vendor is not None and package_id.author != vendor:
                    click.echo(
                        "Skipping the hash update for {} of type {}".format(
                            package_path.name, package_id.package_type
                        )
                    )
                    continue
                package_hashes[key] = package_hash
        click.echo("Done!")
    except Exception:  # pylint: disable=broad-except
        traceback.print_exc()
        return_code = 1

    return return_code


@click.group(name="hash")
def hash_group() -> None:
    """Hashing utils."""


@hash_group.command(name="all")
@click.option(
    "--packages-dir",
    type=click.Path(exists=True, dir_okay=True, file_okay=False),
    default=Path("packages/"),
)
@click.option("--vendor", type=str)
@click.option("--no-wrap", is_flag=True)
def generate_all(
    packages_dir: Path,
    vendor: Optional[str],
    no_wrap: bool,
) -> None:
    """Generate IPFS hashes."""
    message = (
        "`aea hash all` command has been deprecated and will be removed on v2.0.0, "
        "please use `aea packages lock` command to perform package hash updates"
    )
    warn(
        message=message,
        category=DeprecationWarning,
        stacklevel=2,
    )
    click.echo(message=message)
    packages_dir = Path(packages_dir).absolute()
    return_code = update_hashes(packages_dir, no_wrap, vendor=vendor)
    sys.exit(return_code)


@hash_group.command(name="one")
@click.argument("path", type=click.Path(exists=True, file_okay=True, dir_okay=True))
@click.option("--no-wrap", is_flag=True)
def hash_file(path: str, no_wrap: bool) -> None:
    """Hash a single file/directory."""

    click.echo(f"Path : {path}")
    click.echo(f"Hash : {IPFSHashOnly.get(path, wrap=not no_wrap)}")


@hash_group.command(name="to-v0")
@click.argument("hash_string")
def to_v0_string(hash_string: str) -> None:
    """Convert hash to CID v0."""
    try:
        v0_hash = to_v0(hash_string)
        click.echo(v0_hash)
    except Exception as e:  # pylint: disable=broad-except
        raise click.ClickException(str(e)) from e


@hash_group.command(name="to-v1")
@click.argument("hash_string")
@click.option("--encoding", type=str, default=DEFAULT_ENCODING)
def to_v1_string(hash_string: str, encoding: str) -> None:
    """Convert hash to CID v1."""
    try:
        v1_hash = to_v1(hash_string, encoding)
        click.echo(v1_hash)
    except Exception as e:  # pylint: disable=broad-except
        raise click.ClickException(str(e)) from e



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for the private key password support."""

from pathlib import Path

import pytest
from click.exceptions import ClickException

from aea.helpers.base import cd
from aea.test_tools.test_cases import AEATestCaseEmpty


class TestKeyEncryption(AEATestCaseEmpty):
    """Test that the command 'aea generate-key' works as expected."""

    @pytest.mark.parametrize("ledger_name", ["fetchai", "ethereum", "cosmos"])
    def test_crypto_plugin(self, ledger_name):
        """Test that the fetch private key is created correctly."""
        with cd(self._get_cwd()):
            plain_file_name = Path(f"{ledger_name}_key")
            encrypted_file_name = Path(f"{ledger_name}_key_encrypted")
            password = "somePwd"  # nosec
            self.invoke("generate-key", ledger_name, str(plain_file_name))
            assert plain_file_name.exists()

            self.invoke(
                "generate-key",
                ledger_name,
                str(encrypted_file_name),
                "--password",
                password,
            )
            assert encrypted_file_name.exists()
            assert len(encrypted_file_name.read_bytes()) != len(
                plain_file_name.read_bytes()
            )

            with pytest.raises(ClickException, match="Error on key.*load.*password"):
                self.invoke("add-key", ledger_name, str(encrypted_file_name))

            with pytest.raises(ClickException, match="Decrypt error! Bad password?"):
                self.invoke(
                    "add-key",
                    ledger_name,
                    str(encrypted_file_name),
                    "--password",
                    "incorrectpassword",
                )
            r = self.invoke(
                "add-key",
                ledger_name,
                str(encrypted_file_name),
                "--password",
                password,
            )
            assert r.exit_code == 0

            r = self.invoke(
                "get-address",
                ledger_name,
                "--password",
                password,
            )
            assert r.exit_code == 0



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the ipfs helper module."""

import tempfile
from pathlib import Path
from tempfile import TemporaryDirectory

import pytest
from aea_cli_ipfs.ipfs_utils import IPFSTool  # type: ignore

from aea.helpers.cid import to_v1
from aea.helpers.ipfs.base import IPFSHashOnly


def test_hash_for_big_file():
    """Check hash is ok for big amount of data with chunks support."""
    VALID_HASH = "QmWt5fanMr2JbiaUAUpyLUL8FegGn95t5tHA6kgobXgWX3"  # from ipfs daemon
    data = b"1" * int(IPFSHashOnly.DEFAULT_CHUNK_SIZE * 1.5)
    my_hash = IPFSHashOnly._generate_hash(data)
    assert my_hash == VALID_HASH


class TestFileHashing:
    """Test file hashing"""

    @pytest.mark.parametrize(
        "wrap, cid_v1, expected_multihash",
        [
            (1, 1, "bafybeicjexomh6l2rb3efmzojmsx2p2gynjzg3eztf4quu6zyepmnisn4e"),
            (1, 0, "QmTGBxU5aqqpeiihQxcWr4xynhqWt23R73Btss2j8r9XcC"),
            (0, 1, "bafybeidydfeznx64wcyut2quu2xfx7sk3afxbgy5y6qfoj7lic6fgk4khq"),
            (0, 0, "QmWRTzpGNWJZoNkLoQVBRjsSk58P9nfpxx4iAU4MJDRxkb"),
        ],
    )
    def test_get_file_hash(self, wrap, cid_v1, expected_multihash):
        """Test IPFSHashOnly.get file hash"""
        with tempfile.TemporaryDirectory() as tmp_dir:
            file = Path(tmp_dir) / "dummy_file.txt"
            file.write_text("dummy_data")
            wrap, cid_v1 = map(bool, (wrap, cid_v1))
            computed_multihash = IPFSHashOnly.get(str(file), wrap=wrap, cid_v1=cid_v1)
            assert computed_multihash == expected_multihash


@pytest.mark.usefixtures("use_ipfs_daemon")
class TestDirectoryHashing:
    """Test recursive directory hashing."""

    def setup_method(
        self,
    ) -> None:
        """Setup test."""

        self.hash_tool = IPFSHashOnly()
        self.ipfs_tool = IPFSTool(addr="/ip4/127.0.0.1/tcp/5001")

    @pytest.mark.parametrize(
        "wrap, cid_v1, expected_multihash",
        [
            (1, 1, "bafybeies5n7nngqpzwzhety4ndktmzfjgwr5pk2ge3wbftqjwymhlzrc7e"),
            (1, 0, "QmYEAWM6jmjffDQNiyVjVWFcx7SRvutnoRuEP6AJKN9apY"),
            (0, 1, "bafybeicjexomh6l2rb3efmzojmsx2p2gynjzg3eztf4quu6zyepmnisn4e"),
            (0, 0, "QmTGBxU5aqqpeiihQxcWr4xynhqWt23R73Btss2j8r9XcC"),
        ],
    )
    def test_get_dir_hash(self, wrap, cid_v1, expected_multihash):
        """Test IPFSHashOnly.get directory hash"""

        with tempfile.TemporaryDirectory() as tmp_dir:
            file = Path(tmp_dir) / "nested" / "dummy_file.txt"
            file.parent.mkdir()
            file.write_text("dummy_data")
            wrap, cid_v1 = map(bool, (wrap, cid_v1))
            computed_multihash = IPFSHashOnly.get(
                str(file.parent), wrap=wrap, cid_v1=cid_v1
            )
            assert computed_multihash == expected_multihash

    def test_depth_0(
        self,
    ) -> None:
        """Test directory with only one file and no child directories."""

        with TemporaryDirectory() as temp_dir:
            Path(temp_dir, "dummy_file.txt").write_text("Hello, World!")

            hash_local = self.hash_tool.get(temp_dir)
            d, hash_daemon, _ = self.ipfs_tool.add(temp_dir)
            hash_daemon = to_v1(hash_daemon)

            assert (
                hash_daemon == hash_local
            ), f"Hash from daemon {hash_daemon} does not match calculated hash {hash_local}\n{d}"

    def test_depth_1(
        self,
    ) -> None:
        """Test directory with only one file and a child directory."""

        with TemporaryDirectory() as temp_dir:
            Path(temp_dir, "dummy_file.txt").write_text("Hello, World!")
            Path(temp_dir, "inner_0").mkdir()
            Path(temp_dir, "inner_0", "dummy_file_inner.txt").write_text("Foo, Bar!")

            hash_local = self.hash_tool.get(temp_dir)
            d, hash_daemon, _ = self.ipfs_tool.add(temp_dir)
            hash_daemon = to_v1(hash_daemon)

            assert (
                hash_daemon == hash_local
            ), f"Hash from daemon {hash_daemon} does not match calculated hash {hash_local}\n{d}"

            Path(temp_dir, "inner_0", "__pycache__").mkdir()
            assert hash_daemon == self.hash_tool.get(temp_dir)

            Path(temp_dir, "inner_0", "dummy.pyc").touch()
            assert hash_daemon == self.hash_tool.get(temp_dir)

    def test_depth_multi(
        self,
    ) -> None:
        """Test directory with only one file and a child directory."""

        with TemporaryDirectory() as temp_dir:
            Path(temp_dir, "dummy_file.txt").write_text("Hello, World!")
            for i in range(3, 7):
                if i % 2:
                    Path(temp_dir, f"inner_{i}").mkdir()
                    Path(
                        temp_dir, f"inner_{i}", f"dummy_file_inner_{i}.txt"
                    ).write_text(f"Foo, Bar! {i}")
                else:
                    Path(temp_dir, f"inner_{i}").mkdir()
                    for j in range(i):
                        Path(
                            temp_dir, f"inner_{i}", f"dummy_file_inner_{i}_{j}.txt"
                        ).write_text(f"Foo, Bar! {i}_{j}")
                        Path(temp_dir, f"inner_{i}", f"inner_{i}_{j}").mkdir()
                        for k in range(i):
                            Path(
                                temp_dir,
                                f"inner_{i}",
                                f"inner_{i}_{j}",
                                f"dummy_file_inner_{i}_{j}_{k}.txt",
                            ).write_text(f"Foo, Bar! {i}_{j}_{k}")

            # file larger then default chunk size
            Path(temp_dir, "dummy_file_large.txt").write_text(
                "1" * int(IPFSHashOnly.DEFAULT_CHUNK_SIZE * 1.5)
            )

            hash_local = self.hash_tool.get(temp_dir)
            _, hash_daemon, _ = self.ipfs_tool.add(temp_dir)
            hash_daemon = to_v1(hash_daemon)

            assert (
                hash_daemon == hash_local
            ), f"Hash from daemon {hash_daemon} does not match calculated hash {hash_local}"

    def test_hash_bytes(self):
        """Test hash bytes."""
        some_bytes = b"there is some bytes"
        local_hash = self.hash_tool.hash_bytes(
            some_bytes, wrap=False, file_name_if_wrap="some", cid_v1=False
        )
        ipfs_hash = self.ipfs_tool.add_bytes(some_bytes)
        assert local_hash == ipfs_hash

        with pytest.raises(ValueError, match="is required if wrap option is True"):
            self.hash_tool.hash_bytes(some_bytes, wrap=True, cid_v1=False)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""A module with context tools of the aea cli."""

import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, cast

from aea.cli.registry.settings import REGISTRY_LOCAL, REGISTRY_TYPES
from aea.cli.utils.loggers import logger
from aea.configurations.base import (
    AgentConfig,
    Dependencies,
    Dependency,
    PackageType,
    PublicId,
    _get_default_configuration_file_name_from_type,
)
from aea.configurations.constants import (
    DEFAULT_AEA_CONFIG_FILE,
    DEFAULT_REGISTRY_NAME,
    VENDOR,
)
from aea.configurations.loader import ConfigLoader
from aea.configurations.pypi import (
    is_satisfiable,
    is_simple_dep,
    merge_dependencies_list,
    to_set_specifier,
)
from aea.exceptions import AEAException
from aea.helpers.io import open_file


class Context:
    """A class to keep configuration of the cli tool."""

    agent_config: AgentConfig

    _registry_type: Optional[str]

    def __init__(self, cwd: str, verbosity: str, registry_path: Optional[str]) -> None:
        """Init the context."""
        self.config = dict()  # type: Dict
        self.cwd = cwd
        self.verbosity = verbosity
        self.clean_paths: List = []
        self._registry_path = registry_path
        self._registry_type = None

    @property
    def registry_type(
        self,
    ) -> str:
        """Returns registry type to be used for session"""
        if self._registry_type is None:
            logger.warning("Registry not set, returning local as registry type")
            return REGISTRY_LOCAL

        return self._registry_type

    @registry_type.setter
    def registry_type(self, value: str) -> None:
        """Set registry value."""
        if value is not None:
            if value not in REGISTRY_TYPES:
                raise ValueError(
                    f"{value} not allowed as registry type; Allowed registry types: {REGISTRY_TYPES}"
                )

            self._registry_type = value

    @property
    def registry_path(self) -> str:
        """Get registry path specified or from config or default one with check is it present."""
        # registry path is provided or in config or default
        if self._registry_path:
            registry_path = Path(self._registry_path)
            if not (registry_path.exists() and registry_path.is_dir()):
                raise ValueError(
                    f"Registry path directory provided ({self._registry_path}) can not be found. Current work dir is {self.cwd}"
                )
            return str(registry_path)

        registry_path = (Path(self.cwd) / DEFAULT_REGISTRY_NAME).absolute()
        if registry_path.is_dir():
            return str(registry_path)
        registry_path = (Path(self.cwd) / ".." / DEFAULT_REGISTRY_NAME).absolute()
        if registry_path.is_dir():
            return str(registry_path)
        raise ValueError(
            f"Registry path not provided and local registry `{DEFAULT_REGISTRY_NAME}` not found in current ({self.cwd}) and parent directory."
        )

    @property
    def skip_aea_validation(self) -> bool:
        """
        Get the 'skip_aea_validation' flag.

        If true, validation of the AEA version for loaded configuration
        file is skipped.

        :return: the 'skip_aea_validation'
        """
        return self.config.get("skip_aea_validation", True)

    @property
    def agent_loader(self) -> ConfigLoader:
        """Get the agent loader."""
        return ConfigLoader.from_configuration_type(
            PackageType.AGENT, skip_aea_validation=self.skip_aea_validation
        )

    @property
    def protocol_loader(self) -> ConfigLoader:
        """Get the protocol loader."""
        return ConfigLoader.from_configuration_type(
            PackageType.PROTOCOL, skip_aea_validation=self.skip_aea_validation
        )

    @property
    def connection_loader(self) -> ConfigLoader:
        """Get the connection loader."""
        return ConfigLoader.from_configuration_type(
            PackageType.CONNECTION, skip_aea_validation=self.skip_aea_validation
        )

    @property
    def skill_loader(self) -> ConfigLoader:
        """Get the skill loader."""
        return ConfigLoader.from_configuration_type(
            PackageType.SKILL, skip_aea_validation=self.skip_aea_validation
        )

    @property
    def contract_loader(self) -> ConfigLoader:
        """Get the contract loader."""
        return ConfigLoader.from_configuration_type(
            PackageType.CONTRACT, skip_aea_validation=self.skip_aea_validation
        )

    @property
    def custom_loader(self) -> ConfigLoader:
        """Get the custom loader."""
        return ConfigLoader.from_configuration_type(
            PackageType.CUSTOM, skip_aea_validation=self.skip_aea_validation
        )

    def set_config(self, key: str, value: Any) -> None:
        """
        Set a config.

        :param key: the key for the configuration.
        :param value: the value associated with the key.
        """
        self.config[key] = value
        logger.debug("  config[{}] = {}".format(key, value))

    @staticmethod
    def _get_item_dependencies(item_type: str, public_id: PublicId) -> Dependencies:
        """Get the dependencies from item type and public id."""
        item_type_plural = item_type + "s"
        default_config_file_name = _get_default_configuration_file_name_from_type(
            item_type
        )
        path = Path(
            VENDOR,
            public_id.author,
            item_type_plural,
            public_id.name,
            default_config_file_name,
        )
        if not path.exists():
            path = Path(item_type_plural, public_id.name, default_config_file_name)
        config_loader = ConfigLoader.from_configuration_type(item_type)
        with open_file(path) as fp:
            config = config_loader.load(fp)
        deps = cast(Dependencies, config.dependencies)
        return deps

    @staticmethod
    def _find_unsatisfiable_dependencies(dependencies: Dependencies) -> Dependencies:
        """
        Find unsatisfiable dependencies.

        It only checks among 'simple' dependencies (i.e. if it has no field specified,
        or only the 'version' field set.)

        :param dependencies: the dependencies to check.
        :return: the unsatisfiable dependencies.
        """
        return {
            name: dep
            for name, dep in dependencies.items()
            if is_simple_dep(dep) and not is_satisfiable(to_set_specifier(dep))
        }

    def _get_dependencies_by_item_type(self, item_type: PackageType) -> Dependencies:
        """Get the dependencies from item type and public id."""
        if item_type == PackageType.AGENT:
            return self.agent_config.dependencies
        dependency_to_package: Dict[str, List[Tuple[PublicId, Dependency]]] = {}
        dependencies = []
        for item_id in getattr(self.agent_config, item_type.to_plural()):
            package_dependencies = self._get_item_dependencies(item_type.value, item_id)
            dependencies += [package_dependencies]
            for dep, spec in package_dependencies.items():
                if dep not in dependency_to_package:
                    dependency_to_package[dep] = []
                dependency_to_package[dep].append((item_id, spec))

        merged_dependencies = merge_dependencies_list(*dependencies)
        unsat_dependencies = self._find_unsatisfiable_dependencies(merged_dependencies)
        if len(unsat_dependencies) > 0:
            error = f"Error while merging dependencies for {item_type.to_plural()}"
            error += "; Joint version specifier is unsatisfiable for following dependencies:\n"
            error += "======================================\n"
            for name, spec in unsat_dependencies.items():
                error += f"Dependency: {name}\n"
                error += f"Specifier: {to_set_specifier(spec)}\n"
                error += "Packages containing dependency: \n"
                for package, dep_spec in dependency_to_package[name]:
                    error += f"  - {package.without_hash()}: {dep_spec.get_pip_install_args()[0]}\n"
                error += "======================================\n"

            raise AEAException(error[:-1])
        return merged_dependencies

    def get_dependencies(
        self,
        extra_dependencies: Optional[Tuple[Dependency]] = None,
    ) -> Dependencies:
        """
        Aggregate the dependencies from every component.

        :param extra_dependencies: List of the extra dependencies to use, if the
                                extra dependencies and agent dependencies have conflicts
                                the packages from extra dependencies list will be prefered
                                over the agent dependencies
        :return: a list of dependency version specification. e.g. ["web3 >= 1.0.0"]
        """
        dependencies: Dependencies = {}

        def _update_dependencies(updates: Dependencies) -> None:
            """Update dependencies."""
            for dep, spec in updates.items():
                if dep in dependencies and dependencies[dep] != spec:
                    logger.debug(
                        f"`{dependencies[dep].get_pip_install_args()}` "
                        f"will be overridden by {spec.get_pip_install_args()}"
                    )
                dependencies[dep] = spec

        for item_type in (
            PackageType.PROTOCOL,
            PackageType.CONTRACT,
            PackageType.CUSTOM,
            PackageType.CONNECTION,
            PackageType.SKILL,
            PackageType.AGENT,
        ):
            logger.debug(f"Loading {item_type.value} dependencies")
            type_deps = self._get_dependencies_by_item_type(item_type)
            _update_dependencies(type_deps)

        if extra_dependencies is not None and len(extra_dependencies) > 0:
            logger.debug("Loading extra dependencies")
            type_deps = {spec.name: spec for spec in extra_dependencies}
            _update_dependencies(type_deps)

        return dependencies

    def dump_agent_config(self) -> None:
        """Dump the current agent configuration."""
        with open(os.path.join(self.cwd, DEFAULT_AEA_CONFIG_FILE), "w") as f:
            self.agent_loader.dump(self.agent_config, f)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

# pylint: disable=no-member

"""Python package extending the default open-aea ethereum ledger plugin to add support for flashbots."""

import logging
from typing import Any, Dict, List, Optional, Tuple, Union, cast
from uuid import uuid4

from aea_ledger_ethereum import DEFAULT_ADDRESS, EthereumApi, EthereumCrypto
from eth_account import Account
from eth_account.signers.local import LocalAccount
from flashbots import FlashbotProvider, Flashbots, FlashbotsMiddlewareBuilder
from flashbots.flashbots import FlashbotsBundleResponse
from flashbots.types import FlashbotsBundleRawTx, FlashbotsBundleTx
from hexbytes import HexBytes
from web3 import HTTPProvider, Web3
from web3._utils.module import attach_modules
from web3.exceptions import TransactionNotFound

from aea.common import JSONLike
from aea.helpers.base import try_decorator

_default_logger = logging.getLogger("aea.ledger_apis.ethereum_flashbots")

_ETHEREUM_FLASHBOTS = "ethereum_flashbots"

_TARGET_BLOCKS = "target_blocks"

_DEFAULT_TARGET_BLOCKS = 25

_RAISE_ON_FAILED_SIMULATION = "raise_on_failed_simulation"

_USE_ALL_BUILDERS = "use_all_builders"


def multiple_flashbots_builders(
    signature_account: LocalAccount,
    builders: List[Tuple[str, str]],
    rpc_endpoint: str = DEFAULT_ADDRESS,
) -> Dict[str, Web3]:
    """Setup multiple flashbots providers."""
    builder_to_instance = {}
    for builder_name, endpoint_uri in builders:
        flashbots_provider = FlashbotProvider(signature_account, endpoint_uri)
        w3 = Web3(HTTPProvider(endpoint_uri=rpc_endpoint))
        flash_middleware = FlashbotsMiddlewareBuilder.build(flashbots_provider)
        w3.middleware_onion.add(flash_middleware)
        # attach modules to add the new namespace commands
        attach_modules(w3, {"flashbots": (Flashbots,)})
        builder_to_instance[builder_name] = w3
    return builder_to_instance


class EthereumFlashbotApi(EthereumApi):
    """Class to interact with the Ethereum Web3 APIs."""

    identifier = _ETHEREUM_FLASHBOTS

    def __init__(self, **kwargs: Any):
        """
        Initialize the Ethereum API.

        :param kwargs: the keyword arguments.
        """
        rpc_endpoint = kwargs.get("address", DEFAULT_ADDRESS)
        super().__init__(**kwargs)
        authentication_private_key = kwargs.pop("authentication_private_key", None)
        authentication_account = (
            Account.create()  # pylint: disable=no-value-for-parameter
            if authentication_private_key is None
            else Account.from_key(  # pylint: disable=no-value-for-parameter
                private_key=authentication_private_key
            )
        )
        flashbots_builders = kwargs.pop("flashbots_builders", None)
        if (
            flashbots_builders is None
            or not isinstance(flashbots_builders, list)
            or len(flashbots_builders) == 0
        ):
            raise ValueError(
                "flashbots_builders: List[Tuple[str, str]] must be provided."
            )
        # use the first builder as default
        self._default_flashbots_builder_name = flashbots_builders[0][0]
        self._builder_to_web3 = multiple_flashbots_builders(
            authentication_account,
            flashbots_builders,
            rpc_endpoint,
        )

    @property
    def flashbots(self) -> Flashbots:
        """Get the flashbots Web3 module."""
        # use the first builder as default
        builder_name = self._default_flashbots_builder_name
        flashbots_module = getattr(
            self._builder_to_web3[builder_name], "flashbots", None
        )
        if flashbots_module is None:  # pragma: nocover
            raise ValueError(
                f"Flashbots-{builder_name} have not been registered as a Web3 module."
            )
        return cast(Flashbots, flashbots_module)

    def send_to_all_builders(
        self,
        bundle: List[FlashbotsBundleRawTx],
        target_block: int,
        opts: Dict[str, Any],
    ) -> Dict[str, FlashbotsBundleResponse]:
        """Send the transaction to multiple builders."""
        builder_to_response = {}
        for builder_name, web3 in self._builder_to_web3.items():
            flashbots_module = getattr(web3, "flashbots", None)
            if flashbots_module is None:  # pragma: nocover
                raise ValueError(
                    f"Flashbots {builder_name} have not been registered as a Web3 module."
                )
            response = flashbots_module.send_bundle(bundle, target_block, opts=opts)
            _default_logger.info(
                f"Flashbots-{builder_name} send_bundle response: {response}"
            )
            builder_to_response[builder_name] = response
        return builder_to_response

    @staticmethod
    def bundle_transactions(
        signed_transactions: List[JSONLike],
    ) -> List[FlashbotsBundleRawTx]:
        """Bundle transactions."""
        return [
            FlashbotsBundleRawTx(
                signed_transaction=HexBytes(signed_transaction.get("raw_transaction"))
            )
            for signed_transaction in signed_transactions
        ]

    def simulate(
        self,
        bundle: List[Union[FlashbotsBundleTx, FlashbotsBundleRawTx]],
        target_block: Optional[int],
    ) -> bool:
        """
        Simulate a bundle.

        1. Simulate the bundle in a try catch block.
        2. Return True if simulation went through, or False if something went wrong.

        :param bundle: the bundle to simulate.
        :param target_block: the target block for the transaction, the current block if not provided.
        :return: True if the simulation went through, False otherwise.
        """
        _default_logger.debug(f"Simulating bundle: {bundle}")
        try:
            simulation_response = self.flashbots.simulate(bundle, target_block)
            _default_logger.info(
                f"Flashbots simulation response: {simulation_response}"
            )
            _default_logger.debug(f"Simulation successful for bundle {bundle}.")
            return True
        except Exception as e:  # pylint: disable=broad-except
            _default_logger.warning(f"Simulation failed for bundle {bundle}: {e}")
        return False

    def send_bundle(
        self,
        bundle: List[Union[FlashbotsBundleTx, FlashbotsBundleRawTx]],
        target_blocks: List[int],
        raise_on_failed_simulation: bool = False,
        use_all_builders: bool = False,
    ) -> Optional[List[str]]:
        """
        Send a bundle.

        1. Simulate the bundle.
        2. Send the bundle in a try catch block.
        3. Wait for the response. If successful, go to step 4.
         If current block number is less than the maximum target block number, go to step 1.
        4. Return the transaction digests if the transactions went through, or None if something went wrong.

        :param bundle: the signed transactions to bundle together and send.
        :param target_blocks: the target blocks for the transactions.
        :param raise_on_failed_simulation: whether to raise an exception if the simulation fails.
        :param use_all_builders: whether to send the bundle to all builders.
        :return: the transaction digest if the transaction went through, None otherwise.
        """
        for target_block in target_blocks:
            current_block = self.api.eth.block_number
            if current_block >= target_block:
                # we can only target future blocks
                _default_logger.debug(
                    f"Current block {current_block} >= target block {target_block}"
                )
                continue
            # we simulate the bundle against the current block
            if not self.simulate(bundle, current_block):
                msg = f"Simulation failed for bundle {bundle} on block {current_block}."
                if raise_on_failed_simulation:
                    raise ValueError(msg)
                _default_logger.warning(msg)
                continue

            replacement_uuid = str(uuid4())
            _default_logger.debug(
                f"Sending bundle {bundle} with replacement_uuid {replacement_uuid} targeting block {target_block}"
            )
            # we try to send the bundle on the target block, which MUST be greater than the current block
            builder_to_response = {}
            if use_all_builders:
                builder_to_response = self.send_to_all_builders(
                    bundle, target_block, opts={"replacementUuid": replacement_uuid}
                )
            else:
                response = self.flashbots.send_bundle(
                    bundle, target_block, opts={"replacementUuid": replacement_uuid}
                )
                builder_to_response[self._default_flashbots_builder_name] = response
            _default_logger.debug(
                f"Response from bundle with replacement uuid {replacement_uuid}: {builder_to_response}"
            )

            default_builder_response = builder_to_response[
                self._default_flashbots_builder_name
            ]
            # all builders target the same block, so we can just wait for the default builder
            default_builder_response.wait()
            try:
                receipts = default_builder_response.receipts()
                tx_hashes = [tx["hash"].hex() for tx in default_builder_response.bundle]
                _default_logger.debug(
                    f"Bundle with replacement uuid {replacement_uuid} was mined in block {receipts[0]['blockNumber']}"
                    f"Tx hashes: {tx_hashes}"
                )
                return tx_hashes
            except TransactionNotFound:
                # get & log the bundle stats in case it was not included in the block
                stats = self.flashbots.get_bundle_stats_v2(
                    self.api.toHex(default_builder_response.bundle_hash()), target_block
                )
                _default_logger.info(
                    f"Bundle with replacement uuid {replacement_uuid} not found in block {target_block}. "
                    f"bundle stats: {stats}"
                )
                # essentially a no-op but it shows that the function works
                cancel_res = self.flashbots.cancel_bundles(replacement_uuid)
                _default_logger.debug(
                    f"Response from canceling bundle with replacement uuid {replacement_uuid}: {cancel_res}"
                )
        return None

    def _get_next_blocks(self, num_blocks: int = _DEFAULT_TARGET_BLOCKS) -> List[int]:
        """
        Get the next blocks.

        :param num_blocks: the number of blocks to get.
        :return: the next blocks.
        """
        current_block = self.api.eth.block_number
        return list(range(current_block, current_block + num_blocks))

    @try_decorator("Unable to send transactions: {}", logger_method="warning")
    def _try_send_signed_transactions(
        self, signed_transactions: List[JSONLike], **_kwargs: Any
    ) -> Optional[List[str]]:
        """
        Try sending a bundle of transactions.

        :param signed_transactions: the raw signed transactions to bundle together and send.
        :param _kwargs: the keyword arguments. Possible kwargs are:
            `raise_on_try`: bool flag specifying whether the method will raise or log on error (used by `try_decorator`)
            `target_blocks`: the target blocks for the transactions.
            `raise_on_failed_simulation`: whether to raise an exception if the simulation fails.
            `use_all_builders`: whether to send the bundle to all builders.
        :return: the transaction digest if the transactions went through, None otherwise.
        """
        bundle = self.bundle_transactions(signed_transactions)
        target_blocks = _kwargs.get(_TARGET_BLOCKS, self._get_next_blocks())
        raise_on_failed_simulation = _kwargs.get(_RAISE_ON_FAILED_SIMULATION, False)
        use_all_builders = _kwargs.get(_USE_ALL_BUILDERS, False)
        tx_hashes = self.send_bundle(
            bundle, target_blocks, raise_on_failed_simulation, use_all_builders
        )
        return tx_hashes

    def send_signed_transactions(
        self,
        signed_transactions: List[JSONLike],
        raise_on_try: bool = False,
        **kwargs: Any,
    ) -> Optional[List[str]]:
        """
        Simulate and send a bundle of transactions.

        :param signed_transactions: the raw signed transactions to bundle together and send.
        :param raise_on_try: whether to raise an exception if the transaction is not successful.
        :param kwargs: the keyword arguments.
        :return: the transaction digest if the transactions went through, None otherwise.
        """
        tx_hashes = self._try_send_signed_transactions(
            signed_transactions,
            **kwargs,
            raise_on_try=raise_on_try,
        )
        return tx_hashes


class EthereumFlashbotCrypto(EthereumCrypto):
    """Class wrapping the Account Generation from Ethereum ledger."""

    identifier = _ETHEREUM_FLASHBOTS



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This test module contains tests for P2PLibp2p connection."""

# pylint: skip-file

import asyncio
from itertools import permutations
from unittest import mock
from unittest.mock import Mock, call

import pytest

from aea.mail.base import Empty

from packages.valory.connections.p2p_libp2p.connection import NodeClient
from packages.valory.connections.p2p_libp2p.tests.base import libp2p_log_on_failure_all
from packages.valory.connections.test_libp2p.tests.base import (
    BaseP2PLibp2pTest,
    TIMEOUT,
    _make_libp2p_connection,
)

DEFAULT_NET_SIZE = 4


@pytest.mark.asyncio
@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionConnectDisconnect(BaseP2PLibp2pTest):
    """Test that connection is established and torn down correctly"""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("crypto", BaseP2PLibp2pTest._cryptos)
    async def test_p2p_libp2p_connection_connect_disconnect(self, crypto) -> None:
        """Test connect then disconnect."""

        connection = _make_libp2p_connection(agent_key=crypto)
        assert connection.is_connected is False
        try:
            await connection.connect()
            assert connection.is_connected is True
        except Exception as e:
            await connection.disconnect()
            raise e

        await connection.disconnect()
        assert connection.is_connected is False


@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionEchoEnvelope(BaseP2PLibp2pTest):
    """Test that connection will route envelope to destination"""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()

        cls.connection1 = cls.make_connection(agent_key=cls.libp2p_crypto)
        genesis_peer = cls.connection1.node.multiaddrs[0]
        cls.connection2 = cls.make_connection(
            entry_peers=[genesis_peer],
            agent_key=cls.default_crypto,
        )

    def test_connection_is_established(self) -> None:
        """Test connection established."""
        assert self.all_connected

    def test_envelope_routed(self) -> None:
        """Test envelope routed."""

        sender = self.connection1.address
        to = self.connection2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[0].put(envelope)
        delivered_envelope = self.multiplexers[1].get(block=True, timeout=TIMEOUT)

        assert delivered_envelope is not None
        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

    def test_envelope_echoed_back(self) -> None:
        """Test envelope echoed back."""

        sender = self.connection1.address
        to = self.connection2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[0].put(envelope)
        delivered_envelope = self.multiplexers[1].get(block=True, timeout=TIMEOUT)

        assert delivered_envelope is not None
        delivered_envelope.to = sender
        delivered_envelope.sender = to

        self.multiplexers[1].put(delivered_envelope)
        echoed_envelope = self.multiplexers[0].get(block=True, timeout=TIMEOUT)

        assert echoed_envelope is not None
        assert self.sent_is_echoed_envelope(envelope, echoed_envelope)


@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionRouting(BaseP2PLibp2pTest):
    """Test that libp2p node will reliably route envelopes in a local network"""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()

        connection_genesis = cls.make_connection()
        genesis_peer = connection_genesis.node.multiaddrs[0]
        for _ in range(DEFAULT_NET_SIZE):
            cls.make_connection(entry_peers=[genesis_peer])

    def test_connection_is_established(self) -> None:
        """Test connection established."""
        assert self.all_connected

    def test_star_routing_connectivity(self) -> None:
        """Test star routing connectivity."""

        addrs = [c.address for m in self.multiplexers for c in m.connections]
        nodes = range(len(self.multiplexers))
        for u, v in permutations(nodes, 2):
            envelope = self.enveloped_default_message(to=addrs[v], sender=addrs[u])
            self.multiplexers[u].put(envelope)
            delivered_envelope = self.multiplexers[v].get(block=True, timeout=TIMEOUT)
            assert delivered_envelope is not None
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)


@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionEchoEnvelopeRelayOneDHTNode(BaseP2PLibp2pTest):
    """Test that connection will route envelope to destination using the same relay node"""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()

        relay = cls.make_connection()
        relay_peer = relay.node.multiaddrs[0]

        cls.connection1 = cls.make_connection(relay=False, entry_peers=[relay_peer])
        cls.connection2 = cls.make_connection(entry_peers=[relay_peer])

    def test_connection_is_established(self) -> None:
        """Test connection established."""
        assert self.all_connected

    def test_envelope_routed(self) -> None:
        """Test envelope routed."""

        sender = self.connection1.address
        to = self.connection2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[1].put(envelope)
        delivered_envelope = self.multiplexers[2].get(block=True, timeout=TIMEOUT)

        assert delivered_envelope is not None
        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)

    def test_envelope_echoed_back(self) -> None:
        """Test envelope echoed back."""

        sender = self.connection1.address
        to = self.connection2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        self.multiplexers[1].put(envelope)
        delivered_envelope = self.multiplexers[2].get(block=True, timeout=TIMEOUT)

        assert delivered_envelope is not None
        delivered_envelope.to = sender
        delivered_envelope.sender = to

        self.multiplexers[2].put(delivered_envelope)
        echoed_envelope = self.multiplexers[1].get(block=True, timeout=TIMEOUT)

        assert echoed_envelope is not None
        self.sent_is_echoed_envelope(envelope, echoed_envelope)


@libp2p_log_on_failure_all
class TestP2PLibp2pConnectionRoutingRelayTwoDHTNodes(BaseP2PLibp2pTest):
    """Test DHT network routing of envelopes from relay/non-relay to relay/non-relay nodes"""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()

        cls.connection1 = cls.make_connection()
        relay_peer_1 = cls.connection1.node.multiaddrs[0]
        cls.connection2 = cls.make_connection(entry_peers=[relay_peer_1])
        relay_peer_2 = cls.connection2.node.multiaddrs[0]

        for _ in range(DEFAULT_NET_SIZE // 2 + 1):
            cls.make_connection(entry_peers=[relay_peer_1])
            cls.make_connection(entry_peers=[relay_peer_2])

    def test_connection_is_established(self) -> None:
        """Test connection established."""
        assert self.all_connected

    def test_star_routing_connectivity(self) -> None:
        """Test star routing connectivity."""

        addrs = [c.address for m in self.multiplexers for c in m.connections]
        nodes = range(len(self.multiplexers))
        for u, v in permutations(nodes, 2):
            envelope = self.enveloped_default_message(to=addrs[v], sender=addrs[u])
            self.multiplexers[u].put(envelope)
            delivered_envelope = self.multiplexers[v].get(block=True, timeout=TIMEOUT)
            assert delivered_envelope is not None
            assert self.sent_is_delivered_envelope(envelope, delivered_envelope)


@pytest.mark.asyncio
class TestP2PLibp2pNodeRestart(BaseP2PLibp2pTest):
    """Test node restart."""

    @pytest.mark.asyncio
    @pytest.mark.parametrize("crypto", BaseP2PLibp2pTest._cryptos)
    async def test_node_restart(self, crypto) -> None:
        """Test node restart works."""

        connection = _make_libp2p_connection(agent_key=crypto)
        try:
            await connection.node.start()
            pipe = connection.node.pipe
            assert pipe is not None
            await connection.node.restart()
            new_pipe = connection.node.pipe
            assert new_pipe is not None
            assert new_pipe is not pipe
        finally:
            await connection.node.stop()


@libp2p_log_on_failure_all
class BaseTestP2PLibp2p(BaseP2PLibp2pTest):
    """Base test class for p2p libp2p tests with two peers."""

    @classmethod
    def setup_class(cls) -> None:
        """Set the test up"""
        super().setup_class()

        cls.connection1 = cls.make_connection(agent_key=cls.libp2p_crypto)
        genesis_peer = cls.connection1.node.multiaddrs[0]
        cls.connection2 = cls.make_connection(
            entry_peers=[genesis_peer],
            agent_key=cls.default_crypto,
        )


@libp2p_log_on_failure_all
class TestP2PLibp2PSendEnvelope(BaseTestP2PLibp2p):
    """Test that connection will send envelope with error, and that reconnection fixes it."""

    def test_connection_is_established(self) -> None:
        """Test connection established."""
        assert self.all_connected

    def test_envelope_routed(self) -> None:
        """Test envelope routed."""

        sender = self.connection2.address
        to = self.connection1.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        # cause failure on send
        # note: we don't mock the genesis peer.
        with mock.patch.object(
            self.connection2.logger, "exception"
        ) as _mock_logger, mock.patch.object(
            self.connection2.node.pipe, "write", side_effect=Exception("some error")
        ):
            self.multiplexers[1].put(envelope)
            delivered_envelope = self.multiplexers[0].get(block=True, timeout=20)
            _mock_logger.assert_called_with(
                "Failed to send after pipe reconnect. Exception: some error. Try recover connection to node and send again."
            )

        assert delivered_envelope is not None
        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)


@libp2p_log_on_failure_all
class TestP2PLibp2PReceiveEnvelope(BaseTestP2PLibp2p):
    """Test that connection will receive envelope with error, and that reconnection fixes it."""

    def test_envelope_routed(self) -> None:
        """Test envelope routed."""

        sender = self.connection1.address
        to = self.connection2.address
        envelope = self.enveloped_default_message(to=to, sender=sender)

        # make the reception fail
        with mock.patch.object(
            self.connection2.logger, "exception"
        ) as _mock_logger, mock.patch.object(
            self.connection2.node.pipe, "read", side_effect=Exception("some error")
        ):
            self.multiplexers[0].put(envelope)
            delivered_envelope = self.multiplexers[1].get(block=True, timeout=20)
            expected = "Failed to read. Exception: some error. Try reconnect to node and read again."
            _mock_logger.assert_has_calls([call(expected)])

        assert delivered_envelope is not None
        assert self.sent_is_delivered_envelope(envelope, delivered_envelope)


@libp2p_log_on_failure_all
class TestLibp2pEnvelopeOrder(BaseTestP2PLibp2p):
    """
    Test message ordering.

    Test that the order of envelope is the guaranteed to be the same
    when communicating between two peers.
    """

    NB_ENVELOPES = 1000

    def test_burst_order(self) -> None:
        """Test order of envelope burst is guaranteed on receiving end."""

        sender = self.connection1.address
        to = self.connection2.address
        sent_envelopes = [
            self.enveloped_default_message(to, sender) for _ in range(self.NB_ENVELOPES)
        ]

        for envelope in sent_envelopes:
            envelope.message = envelope.message_bytes  # encode
            self.multiplexers[0].put(envelope)

        received_envelopes = []
        for _ in range(self.NB_ENVELOPES):
            envelope = self.multiplexers[1].get(block=True, timeout=3)
            received_envelopes.append(envelope)

        # test no new message is "created"
        with pytest.raises(Empty):
            self.multiplexers[1].get(block=True, timeout=1)

        assert len(sent_envelopes) == len(received_envelopes)
        for expected, actual in zip(sent_envelopes, received_envelopes):
            assert expected.message == actual.message


@pytest.mark.asyncio
async def test_node_client_pipe_connect() -> None:
    """Test pipe.connect called on NodeClient.connect."""
    f = asyncio.Future()
    f.set_result(None)
    pipe = Mock()
    pipe.connect.return_value = f
    node_client = NodeClient(pipe, Mock())
    await node_client.connect()
    pipe.connect.assert_called_once()


@pytest.mark.asyncio
async def test_write_acn_error() -> None:
    """Test pipe.connect called on NodeClient.connect."""
    f = asyncio.Future()
    f.set_result(None)
    pipe = Mock()
    pipe.write.return_value = f
    node_client = NodeClient(pipe, Mock())
    await node_client.write_acn_status_error("some message")
    pipe.write.assert_called_once()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the strategy class of the erc1155_client skill."""

# pylint: skip-file

from aea.helpers.search.models import Constraint, ConstraintType, Query

from packages.fetchai.skills.erc1155_client.strategy import (
    CONTRACT_ID,
    SIMPLE_SERVICE_MODEL,
)
from packages.fetchai.skills.erc1155_client.tests.intermediate_class import (
    ERC1155ClientTestCase,
)


class TestStrategy(ERC1155ClientTestCase):
    """Test Strategy of erc1155_client."""

    def test_properties(self):
        """Test the properties of Strategy class."""
        assert self.strategy.ledger_id == self.skill.skill_context.default_ledger_id
        assert self.strategy.contract_id == str(CONTRACT_ID)

    def test_get_location_and_service_query(self):
        """Test the get_location_and_service_query method of the Strategy class."""
        query = self.strategy.get_location_and_service_query()

        assert type(query) is Query
        assert len(query.constraints) == 2
        assert query.model is None

        location_constraint = Constraint(
            "location",
            ConstraintType(
                "distance", (self.strategy._agent_location, self.search_radius)
            ),
        )
        assert query.constraints[0] == location_constraint

        service_key_constraint = Constraint(
            self.search_query["search_key"],
            ConstraintType(
                self.search_query["constraint_type"],
                self.search_query["search_value"],
            ),
        )
        assert query.constraints[1] == service_key_constraint

    def test_get_service_query(self):
        """Test the get_service_query method of the Strategy class."""
        query = self.strategy.get_service_query()

        assert type(query) is Query
        assert len(query.constraints) == 1

        assert query.model == SIMPLE_SERVICE_MODEL

        service_key_constraint = Constraint(
            self.search_query["search_key"],
            ConstraintType(
                self.search_query["constraint_type"],
                self.search_query["search_value"],
            ),
        )
        assert query.constraints[0] == service_key_constraint



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests for acn helper module."""

from aea.helpers.acn.uri import Uri


def test_uri():
    """Test URI class"""
    Uri(uri="localhost:9000")
    uri = Uri(host="localhost", port=9000)
    assert str(uri) == "localhost:9000"
    assert uri.host == "localhost"
    assert uri.port == 9000
    Uri()


def test_uri2():
    """Test the uri."""
    Uri(host="127.0.0.1")
    uri = Uri(host="127.0.0.1", port=10000)
    assert str(uri) == "127.0.0.1:10000"
    assert uri.host == "127.0.0.1"
    assert uri.port == 10000



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: base.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\nbase.proto\x12\x0f\x61\x65\x61.base.v0_1_0\x1a\x1cgoogle/protobuf/struct.proto\"\x90\x01\n\x0f\x44ialogueMessage\x12\x12\n\nmessage_id\x18\x01 \x01(\x05\x12\"\n\x1a\x64ialogue_starter_reference\x18\x02 \x01(\t\x12$\n\x1c\x64ialogue_responder_reference\x18\x03 \x01(\t\x12\x0e\n\x06target\x18\x04 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x05 \x01(\x0c\"{\n\x07Message\x12\'\n\x04\x62ody\x18\x01 \x01(\x0b\x32\x17.google.protobuf.StructH\x00\x12<\n\x10\x64ialogue_message\x18\x02 \x01(\x0b\x32 .aea.base.v0_1_0.DialogueMessageH\x00\x42\t\n\x07message\"Y\n\x08\x45nvelope\x12\n\n\x02to\x18\x01 \x01(\t\x12\x0e\n\x06sender\x18\x02 \x01(\t\x12\x13\n\x0bprotocol_id\x18\x03 \x01(\t\x12\x0f\n\x07message\x18\x04 \x01(\x0c\x12\x0b\n\x03uri\x18\x05 \x01(\tb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'base_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_DIALOGUEMESSAGE']._serialized_start=62
  _globals['_DIALOGUEMESSAGE']._serialized_end=206
  _globals['_MESSAGE']._serialized_start=208
  _globals['_MESSAGE']._serialized_end=331
  _globals['_ENVELOPE']._serialized_start=333
  _globals['_ENVELOPE']._serialized_end=422
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This package contains the behaviour for the generic buyer skill."""

from typing import Any, List, Optional, Set, cast

from aea.protocols.dialogue.base import DialogueLabel
from aea.skills.behaviours import TickerBehaviour

from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.skills.generic_buyer.dialogues import (
    FipaDialogue,
    LedgerApiDialogue,
    LedgerApiDialogues,
    OefSearchDialogues,
)
from packages.fetchai.skills.generic_buyer.strategy import GenericStrategy
from packages.valory.connections.ledger.connection import (
    PUBLIC_ID as LEDGER_CONNECTION_PUBLIC_ID,
)
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

DEFAULT_MAX_PROCESSING = 120
DEFAULT_TX_INTERVAL = 2.0
DEFAULT_SEARCH_INTERVAL = 5.0
LEDGER_API_ADDRESS = str(LEDGER_CONNECTION_PUBLIC_ID)


class GenericSearchBehaviour(TickerBehaviour):
    """This class implements a search behaviour."""

    def __init__(self, **kwargs: Any):
        """Initialize the search behaviour."""
        search_interval = cast(
            float, kwargs.pop("search_interval", DEFAULT_SEARCH_INTERVAL)
        )
        super().__init__(tick_interval=search_interval, **kwargs)

    def setup(self) -> None:
        """Implement the setup for the behaviour."""
        strategy = cast(GenericStrategy, self.context.strategy)
        if strategy.is_ledger_tx:
            ledger_api_dialogues = cast(
                LedgerApiDialogues, self.context.ledger_api_dialogues
            )
            ledger_api_msg, _ = ledger_api_dialogues.create(
                counterparty=LEDGER_API_ADDRESS,
                performative=LedgerApiMessage.Performative.GET_BALANCE,
                ledger_id=strategy.ledger_id,
                address=cast(str, self.context.agent_addresses.get(strategy.ledger_id)),
            )
            self.context.outbox.put_message(message=ledger_api_msg)
        else:
            strategy.is_searching = True

    def act(self) -> None:
        """Implement the act."""
        strategy = cast(GenericStrategy, self.context.strategy)
        if not strategy.is_searching:
            return
        transaction_behaviour = cast(
            GenericTransactionBehaviour, self.context.behaviours.transaction
        )
        remaining_transactions_count = len(transaction_behaviour.waiting)
        if remaining_transactions_count > 0:
            self.context.logger.info(
                f"Transaction behaviour has {remaining_transactions_count} transactions remaining. Skipping search!"
            )
            return
        strategy.update_search_query_params()
        query = strategy.get_location_and_service_query()
        oef_search_dialogues = cast(
            OefSearchDialogues, self.context.oef_search_dialogues
        )
        oef_search_msg, _ = oef_search_dialogues.create(
            counterparty=self.context.search_service_address,
            performative=OefSearchMessage.Performative.SEARCH_SERVICES,
            query=query,
        )
        self.context.outbox.put_message(message=oef_search_msg)

    def teardown(self) -> None:
        """Implement the task teardown."""


class GenericTransactionBehaviour(TickerBehaviour):
    """A behaviour to sequentially submit transactions to the blockchain."""

    def __init__(self, **kwargs: Any):
        """Initialize the transaction behaviour."""
        tx_interval = cast(
            float, kwargs.pop("transaction_interval", DEFAULT_TX_INTERVAL)
        )
        self.max_processing = cast(
            float, kwargs.pop("max_processing", DEFAULT_MAX_PROCESSING)
        )
        self.processing_time = 0.0
        self.waiting: List[FipaDialogue] = []
        self.processing: Optional[LedgerApiDialogue] = None
        self.timedout: Set[DialogueLabel] = set()
        super().__init__(tick_interval=tx_interval, **kwargs)

    def setup(self) -> None:
        """Setup behaviour."""

    def act(self) -> None:
        """Implement the act."""
        if self.processing is not None:
            if self.processing_time <= self.max_processing:
                # already processing
                self.processing_time += self.tick_interval
                return
            self._timeout_processing()
        if len(self.waiting) == 0:
            # nothing to process
            return
        self._start_processing()

    def _start_processing(self) -> None:
        """Process the next transaction."""
        fipa_dialogue = self.waiting.pop(0)
        self.context.logger.info(
            f"Processing transaction, {len(self.waiting)} transactions remaining"
        )
        ledger_api_dialogues = cast(
            LedgerApiDialogues, self.context.ledger_api_dialogues
        )
        ledger_api_msg, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=LEDGER_API_ADDRESS,
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,
            terms=fipa_dialogue.terms,
        )
        ledger_api_dialogue = cast(LedgerApiDialogue, ledger_api_dialogue)
        ledger_api_dialogue.associated_fipa_dialogue = fipa_dialogue
        self.processing_time = 0.0
        self.processing = ledger_api_dialogue
        self.context.logger.info(
            f"requesting transfer transaction from ledger api for message={ledger_api_msg}..."
        )
        self.context.outbox.put_message(message=ledger_api_msg)

    def teardown(self) -> None:
        """Teardown behaviour."""

        self.processing = None
        self.processing_time = 0.0
        self.timedout.clear()
        self.waiting.clear()

    def _timeout_processing(self) -> None:
        """Timeout processing."""
        if self.processing is None:
            return
        self.timedout.add(self.processing.dialogue_label)
        self.waiting.append(self.processing.associated_fipa_dialogue)
        self.processing_time = 0.0
        self.processing = None

    def finish_processing(self, ledger_api_dialogue: LedgerApiDialogue) -> None:
        """
        Finish processing.

        :param ledger_api_dialogue: the ledger api dialogue
        """
        if self.processing == ledger_api_dialogue:
            self.processing_time = 0.0
            self.processing = None
            return
        if ledger_api_dialogue.dialogue_label not in self.timedout:
            raise ValueError(
                f"Non-matching dialogues in transaction behaviour: {self.processing} and {ledger_api_dialogue}"
            )
        self.timedout.remove(ledger_api_dialogue.dialogue_label)
        self.context.logger.debug(
            f"Timeout dialogue in transaction processing: {ledger_api_dialogue}"
        )
        # don't reset, as another might be processing

    def failed_processing(self, ledger_api_dialogue: LedgerApiDialogue) -> None:
        """
        Failed processing.

        Currently, we retry processing indefinitely.

        :param ledger_api_dialogue: the ledger api dialogue
        """
        self.finish_processing(ledger_api_dialogue)
        self.waiting.append(ledger_api_dialogue.associated_fipa_dialogue)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test if constants are valid."""

from pathlib import Path

from aea.configurations.constants import PACKAGES, SIGNING_PROTOCOL_WITH_HASH
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.package_manager.v1 import PackageManagerV1

from tests.conftest import ROOT_DIR


def test_signing_protocol_hash() -> None:
    """Test signing protocol"""

    package_manager = PackageManagerV1.from_dir(Path(ROOT_DIR, PACKAGES))
    public_id = PublicId.from_str(SIGNING_PROTOCOL_WITH_HASH)
    package_id = PackageId(PackageType.PROTOCOL, public_id)

    assert public_id.hash == package_manager.dev_packages[package_id]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test utilities."""

import os
import shutil
import tempfile
from pathlib import Path
from typing import Type
from unittest import mock

import pytest

from aea.test_tools import utils
from aea.test_tools.utils import copy_class, remove_test_directory, wait_for_condition


def test_wait_for_condition():
    """Test test_wait_for_condition"""

    wait_for_condition(lambda: True)
    with pytest.raises(TimeoutError, match="test error msg"):
        wait_for_condition(lambda: False, error_msg="test error msg")


class TestCopyClass:
    """Test copy_class"""

    cls: Type
    copy_of_cls: Type

    def setup_method(self):
        """Setup"""

        # redefine for each test
        class Parent:
            """Parent"""

            def __init__(self):
                self.x = self._y = None
                self.attr = "attr"
                self.mutable_attr = []

        class Child(Parent):
            """Child"""

            cls_attr = "cls_attr"
            cls_mutable_attr = []

            def __init__(self):
                super().__init__()

            @classmethod
            def cls_method(cls, x):
                cls.x = x

            def method(self, y):
                """F"""
                self._y = y

            @property
            def y(self):
                return self._y

        self.cls = Child
        self.copy_of_cls = copy_class(Child)

    @pytest.mark.parametrize("mutate_original", [False, True])
    def test_cls(self, mutate_original: bool):
        """Test cls"""

        original, replica = self.cls, self.copy_of_cls
        a, b = (original, replica) if mutate_original else (replica, original)

        assert a is not b
        assert a != b
        assert a.cls_attr == b.cls_attr
        assert a.cls_mutable_attr == b.cls_mutable_attr

        # mutate state
        assert not hasattr(a, "x")
        a.cls_method(3)
        assert hasattr(a, "x")
        a.cls_mutable_attr.append(1)
        assert a.cls_mutable_attr
        assert not b.cls_mutable_attr

        # if we set the attribute on the original (parent class),
        # the replicate (child) will also have it.
        a.new_attr = "new_attr"
        a.g = lambda: None
        if not mutate_original:
            assert not hasattr(b, "x")
            assert not hasattr(b, "new_attr")
            assert not hasattr(b, "g")

    @pytest.mark.parametrize("mutate_original", [False, True])
    def test_instance(self, mutate_original: bool):
        """Test instance"""

        original, replica = self.cls(), self.copy_of_cls()
        a, b = (original, replica) if mutate_original else (replica, original)

        assert a is not b
        assert a != b
        assert a.cls_attr == b.cls_attr
        assert a.cls_mutable_attr == b.cls_mutable_attr

        # mutate state
        assert hasattr(a, "x") and hasattr(b, "x")  # set in __init__
        a.cls_method(3)
        assert a.x is b.x is None  # class method does not affect instance

        a.method(3)
        assert a.y == 3
        assert b.y is None

        a.cls_mutable_attr.append(1)
        assert a.cls_mutable_attr
        assert not b.cls_mutable_attr

        # mutating instance attributes in not heritable
        a.new_attr = "new_attr"
        a.g = lambda: None
        assert not hasattr(b, "new_attr")
        assert not hasattr(b, "g")


@pytest.mark.parametrize("path_type", [str, Path])
def test_remove_non_empty_test_directory(path_type):
    """Test remove_test_directory"""

    tmp_dir = path_type(tempfile.TemporaryDirectory().name)
    assert not os.path.exists(tmp_dir)
    shutil.copytree(str(Path(utils.__file__).parent), tmp_dir)
    assert os.path.isdir(tmp_dir)
    assert list(Path(tmp_dir).glob("*"))

    permission = os.stat(tmp_dir).st_mode
    with mock.patch("os.lstat", side_effect=Exception):
        assert not remove_test_directory(tmp_dir)
    assert os.path.exists(tmp_dir)
    assert os.stat(tmp_dir).st_mode == permission

    assert remove_test_directory(tmp_dir)
    assert not os.path.exists(tmp_dir)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains some constants for the p2p libp2p connection."""

import os
import platform

LIBP2P_CERT_NOT_BEFORE = "2026-01-01"
LIBP2P_CERT_NOT_AFTER = "2027-01-01"

LIBP2P_NODE_MODULE_NAME = "libp2p_node"

LIBP2P_NODE_MODULE = str(
    os.path.join(os.path.abspath(os.path.dirname(__file__)), LIBP2P_NODE_MODULE_NAME)
)

if platform.system() == "Windows":  # pragma: nocover
    LIBP2P_NODE_MODULE_NAME += ".exe"

LIBP2P_NODE_DEPS_DOWNLOAD_TIMEOUT = 660  # time to download ~66Mb



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#   Copyright 2018-2021 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Profiler blacklisted types."""

PROFILER_TYPE_BLACK_LIST = {
    "ABCMeta",
    "Add",
    "AddressFamily",
    "AddressInfo",
    "AlertDescription",
    "And",
    "AugLoad",
    "AugStore",
    "BitAnd",
    "BitOr",
    "BitXor",
    "BufferedReader",
    "BufferedWriter",
    "CachedProperty",
    "CodecInfo",
    "Condition",
    "Counter",
    "Del",
    "Div",
    "DynamicClassAttribute",
    "Enum",
    "EnumMeta",
    "Eq",
    "Event",
    "ExtensionFileLoader",
    "FQP_corresponding_FQ_class",
    "FileFinder",
    "FileIO",
    "FloorDiv",
    "Formatter",
    "ForwardRef",
    "Gt",
    "GtE",
    "Handlers",
    "In",
    "IncrementalDecoder",
    "IncrementalEncoder",
    "Invert",
    "Is",
    "IsNot",
    "LShift",
    "Literal",
    "Load",
    "Logger",
    "Lt",
    "LtE",
    "Manager",
    "Marker",
    "MatMult",
    "MessageDescriptor",
    "Mod",
    "ModuleSpec",
    "MovedAttribute",
    "MsgFlag",
    "Mult",
    "NamedTupleMeta",
    "Not",
    "NotEq",
    "NotIn",
    "Op",
    "Options",
    "Or",
    "OrderedDict",
    "Param",
    "Parameter",
    "PercentStyle",
    "PlaceHolder",
    "Pow",
    "Purpose",
    "Quitter",
    "RShift",
    "Regex",
    "RegexFlag",
    "Repr",
    "Requirement",
    "RootLogger",
    "SSLErrorNumber",
    "Sigmasks",
    "Signals",
    "SimpleNamespace",
    "SocketKind",
    "SourceFileLoader",
    "Specifier",
    "SpecifierSet",
    "Store",
    "Sub",
    "TLSVersion",
    "TextIOWrapper",
    "TypeVar",
    "UAdd",
    "USub",
    "Value",
    "Variable",
    "VerifyFlags",
    "VerifyMode",
    "WeakSet",
    "WeakValueDictionary",
    "_ASN1Object",
    "_Environ",
    "_Finder",
    "_GenericAlias",
    "_Helper",
    "_MainThread",
    "_NamedIntConstant",
    "_NamespaceLoader",
    "_NamespacePath",
    "_ParameterKind",
    "_Printer",
    "_ProtocolMeta",
    "_RunningLoop",
    "_SSLMethod",
    "_SendfileMode",
    "_SpecialForm",
    "_StderrHandler",
    "_TLSAlertType",
    "_TLSContentType",
    "_TLSMessageType",
    "_TemplateMetaclass",
    "_TypedDictMeta",
    "_VariadicGenericAlias",
    "_lru_cache_wrapper",
    "_ttinfo",
    "_tuplegetter",
    "builtin_function_or_method",
    "cell",
    "classmethod",
    "classmethod_descriptor",
    "count",
    "curry",
    "cython_function_or_method",
    "date",
    "datetime",
    "deque",
    "dict",
    "frozenset",
    "function",
    "fused_cython_function",
    "getset_descriptor",
    "list",
    "mappingproxy",
    "member_descriptor",
    "method",
    "method-wrapper",
    "method_descriptor",
    "module",
    "partial",
    "property",
    "set",
    "slice",
    "staticmethod",
    "time",
    "timedelta",
    "timezone",
    "tuple",
    "type",
    "tzfile",
    "vectorize",
    "weakref",
    "wrapper_descriptor",
}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Script to generate a markdown package table."""

from pathlib import Path
from typing import Dict

from aea.cli.packages import get_package_manager

COL_WIDTH = 61


def get_packages() -> Dict[str, str]:
    """Get packages."""
    data = get_package_manager(Path("packages").relative_to(".")).json
    if "dev" in data:
        return data["dev"]
    return data


def generate_table() -> None:
    """Generates a markdown table containing a package list"""
    data = get_packages()

    # Table header
    content = (
        f"| {'Package name'.ljust(COL_WIDTH, ' ')} | {'Package hash'.ljust(COL_WIDTH, ' ')} |\n"
        f"| {'-'*COL_WIDTH} | {'-'*COL_WIDTH} |\n"
    )

    # Table rows
    for package, package_hash in data.items():
        package_cell = package.ljust(COL_WIDTH, " ")
        hash_cell = f"`{package_hash}`".ljust(COL_WIDTH, " ")
        content += f"| {package_cell} | {hash_cell} |\n"

    # Write table
    with open(
        Path("docs", "package_list.md"), mode="w", encoding="utf-8"
    ) as packages_list:
        packages_list.write(content)


if __name__ == "__main__":
    generate_table()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2026 Valory AG
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the p2p libp2p connection."""

import asyncio
import logging
import os
import platform
import subprocess  # nosec
import sys
from asyncio import AbstractEventLoop, CancelledError, events
from ipaddress import ip_address
from pathlib import Path
from socket import gethostbyname
from typing import Any, IO, List, Optional, Sequence, cast

from aea.configurations.base import PublicId
from aea.configurations.constants import DEFAULT_LEDGER
from aea.connections.base import Connection, ConnectionStates
from aea.crypto.base import Crypto
from aea.exceptions import enforce
from aea.helpers.acn.agent_record import AgentRecord
from aea.helpers.acn.uri import Uri
from aea.helpers.multiaddr.base import MultiAddr
from aea.helpers.pipe import IPCChannel, TCPSocketChannel
from aea.mail.base import Envelope

from packages.valory.connections.p2p_libp2p.consts import LIBP2P_NODE_MODULE_NAME
from packages.valory.protocols.acn import acn_pb2
from packages.valory.protocols.acn.message import AcnMessage

_default_logger = logging.getLogger("aea.packages.valory.connections.p2p_libp2p")

ACN_CURRENT_VERSION = "0.1.0"

LIBP2P_NODE_LOG_FILE = "libp2p_node.log"

LIBP2P_NODE_ENV_FILE = ".env.libp2p"

LIBP2P_NODE_CLARGS = list()  # type: List[str]


PIPE_CONN_TIMEOUT = 10.0

PUBLIC_ID = PublicId.from_str("valory/p2p_libp2p:0.1.0")

SUPPORTED_LEDGER_IDS = ["fetchai", "cosmos", "ethereum"]

LIBP2P_SUCCESS_MESSAGE = "Peer running in "

POR_DEFAULT_SERVICE_ID = "acn"


def _ip_all_private_or_all_public(addrs: List[str]) -> bool:
    if len(addrs) == 0:
        return True

    is_private = ip_address(gethostbyname(addrs[0])).is_private
    is_loopback = ip_address(gethostbyname(addrs[0])).is_loopback

    for addr in addrs:
        if ip_address(gethostbyname(addr)).is_private != is_private:
            return False  # pragma: nocover
        if ip_address(gethostbyname(addr)).is_loopback != is_loopback:
            return False
    return True


def _golang_module_run(
    path: str,
    name: str,
    args: Sequence[str],
    log_file_desc: IO[str],
    logger: logging.Logger = _default_logger,
) -> subprocess.Popen:
    """
    Runs a built module located at `path`.

    :param path: the path to the go module.
    :param name: the name of the module.
    :param args: the args
    :param log_file_desc: the file descriptor of the log file.
    :param logger: the logger
    :return: subprocess
    """
    cmd = [os.path.join(path, name)]

    cmd.extend(args)

    env = os.environ
    try:
        logger.debug(cmd)
        proc = subprocess.Popen(  # nosec
            cmd,
            cwd=path,
            env=env,
            stdout=log_file_desc,
            stderr=log_file_desc,
            shell=False,
        )
    except Exception as e:
        logger.error(
            "While executing go run . {} at {} : {}".format(path, args, str(e))
        )
        raise e

    return proc


class NodeClient:
    """Client to communicate with node using ipc channel(pipe)."""

    ACN_ACK_TIMEOUT = 5.0

    def __init__(self, pipe: IPCChannel, agent_record: AgentRecord) -> None:
        """Set node client with pipe."""
        self.pipe = pipe
        self.agent_record = agent_record
        self._wait_status: Optional[asyncio.Future] = None

    async def connect(self) -> bool:
        """Connect to node with pipe."""
        return await self.pipe.connect()

    async def send_envelope(self, envelope: Envelope) -> None:
        """Send envelope to node."""
        self._wait_status = asyncio.Future()
        buf = self.make_acn_envelope_message(envelope)
        await self._write(buf)

        status = await self.wait_for_status()

        if status.code != int(AcnMessage.StatusBody.StatusCode.SUCCESS):  # type: ignore  # pylint: disable=no-member
            raise ValueError(  # pragma: nocover
                f"failed to send envelope. got error confirmation: {status.code}"
            )

    async def wait_for_status(self) -> Any:
        """Get status."""
        if self._wait_status is None:  # pragma: nocover
            raise ValueError("value failed!")
        try:
            status = await asyncio.wait_for(
                self._wait_status, timeout=self.ACN_ACK_TIMEOUT
            )
            return status
        except asyncio.TimeoutError:  # pragma: nocover
            if not self._wait_status.done():  # pragma: nocover
                self._wait_status.set_exception(Exception("Timeout"))
            await asyncio.sleep(0)
            raise ValueError("acn status await timeout!")
        finally:
            self._wait_status = None

    @staticmethod
    def make_acn_envelope_message(envelope: Envelope) -> bytes:
        """Make acn message with envelope in."""
        acn_msg = acn_pb2.AcnMessage()  # type: ignore  # pylint: disable=no-member
        performative = acn_pb2.AcnMessage.Aea_Envelope_Performative()  # type: ignore  # pylint: disable=no-member
        performative.envelope = envelope.encode()
        acn_msg.aea_envelope.CopyFrom(performative)  # pylint: disable=no-member
        buf = acn_msg.SerializeToString()
        return buf

    async def read_envelope(self) -> Optional[Envelope]:
        """Read envelope from the node."""
        while True:
            buf = await self._read()

            if not buf:
                return None

            try:
                acn_msg = acn_pb2.AcnMessage()  # type: ignore  # pylint: disable=no-member
                acn_msg.ParseFromString(buf)

            except Exception as e:
                await self.write_acn_status_error(
                    f"Failed to parse acn message {e}",
                    status_code=AcnMessage.StatusBody.StatusCode.ERROR_DECODE,
                )
                raise ValueError(f"Error parsing acn message: {e}") from e

            performative = acn_msg.WhichOneof("performative")
            if performative == "aea_envelope":  # pragma: nocover
                aea_envelope = acn_msg.aea_envelope  # pylint: disable=no-member
                try:
                    envelope = Envelope.decode(aea_envelope.envelope)
                    await self.write_acn_status_ok()
                    return envelope
                except Exception as e:
                    await self.write_acn_status_error(
                        f"Failed to decode envelope: {e}",
                        status_code=AcnMessage.StatusBody.StatusCode.ERROR_DECODE,
                    )
                    raise

            elif performative == "status":
                if self._wait_status is not None:
                    self._wait_status.set_result(
                        acn_msg.status.body  # pylint: disable=no-member
                    )
            else:  # pragma: nocover
                await self.write_acn_status_error(
                    f"Bad acn message {performative}",
                    status_code=AcnMessage.StatusBody.StatusCode.ERROR_UNEXPECTED_PAYLOAD,
                )

    async def write_acn_status_ok(self) -> None:
        """Send acn status ok."""
        acn_msg = acn_pb2.AcnMessage()  # type: ignore  # pylint: disable=no-member
        performative = acn_pb2.AcnMessage.Status_Performative()  # type: ignore  # pylint: disable=no-member
        status = AcnMessage.StatusBody(
            status_code=AcnMessage.StatusBody.StatusCode.SUCCESS, msgs=[]
        )
        AcnMessage.StatusBody.encode(
            performative.body, status  # pylint: disable=no-member
        )
        acn_msg.status.CopyFrom(performative)  # pylint: disable=no-member
        buf = acn_msg.SerializeToString()
        await self._write(buf)

    async def write_acn_status_error(
        self,
        msg: str,
        status_code: AcnMessage.StatusBody.StatusCode = AcnMessage.StatusBody.StatusCode.ERROR_GENERIC,  # type: ignore
    ) -> None:
        """Send acn status error generic."""
        acn_msg = acn_pb2.AcnMessage()  # type: ignore  # pylint: disable=no-member
        performative = acn_pb2.AcnMessage.Status_Performative()  # type: ignore  # pylint: disable=no-member
        status = AcnMessage.StatusBody(status_code=status_code, msgs=[msg])
        AcnMessage.StatusBody.encode(
            performative.body, status  # pylint: disable=no-member
        )
        acn_msg.status.CopyFrom(performative)  # pylint: disable=no-member

        buf = acn_msg.SerializeToString()

        await self._write(buf)

    async def _write(self, data: bytes) -> None:
        """
        Write to the writer stream.

        :param data: data to write to stream
        """
        await self.pipe.write(data)

    async def _read(self) -> Optional[bytes]:
        """
        Read from the reader stream.

        :return: bytes
        """
        return await self.pipe.read()


class Libp2pNode:
    """Libp2p p2p node as a subprocess with named pipes interface."""

    def __init__(  # pylint: disable=too-many-positional-arguments
        self,
        agent_record: AgentRecord,
        key: Crypto,
        module_path: str,
        data_dir: str,
        clargs: Optional[List[str]] = None,
        uri: Optional[Uri] = None,
        public_uri: Optional[Uri] = None,
        delegate_uri: Optional[Uri] = None,
        monitoring_uri: Optional[Uri] = None,
        entry_peers: Optional[Sequence[MultiAddr]] = None,
        log_file: Optional[str] = None,
        env_file: Optional[str] = None,
        logger: logging.Logger = _default_logger,
        peer_registration_delay: Optional[float] = None,
        records_storage_path: Optional[str] = None,
        connection_timeout: Optional[float] = None,
        max_restarts: int = 5,
        mailbox_uri: str = "127.0.0.1:8888",
    ):
        """
        Initialize a p2p libp2p node.

        :param agent_record: the agent proof-of-representation for peer.
        :param key: secp256k1 curve private key.
        :param module_path: the module path.
        :param data_dir: the data directory.
        :param clargs: the command line arguments for the libp2p node
        :param uri: libp2p node ip address and port number in format ipaddress:port.
        :param public_uri: libp2p node public ip address and port number in format ipaddress:port.
        :param delegate_uri: libp2p node delegate service ip address and port number in format ipaddress:port.
        :param monitoring_uri: libp2 node monitoring ip address and port in format ipaddress:port
        :param entry_peers: libp2p entry peers multiaddresses.
        :param log_file: the logfile path for the libp2p node
        :param env_file: the env file path for the exchange of environment variables
        :param logger: the logger.
        :param peer_registration_delay: add artificial delay to agent registration in seconds
        :param records_storage_path: the path where to store the agent records.
        :param connection_timeout: the connection timeout of the node.
        :param max_restarts: amount of node restarts during operation.
        :param mailbox_uri: libp2p mailbox_uri ip address and port number in format ipaddress:port.
        """

        self.record = agent_record
        self.address = self.record.address

        # node id in the p2p network
        self.key = key.private_key
        self.pub = key.public_key

        # node uri
        self.uri = uri if uri is not None else Uri()

        # node public uri, optional
        self.public_uri = public_uri

        # node delegate uri, optional
        self.delegate_uri = delegate_uri

        # node monitoring uri, optional
        self.monitoring_uri = monitoring_uri

        # entry peer
        self.entry_peers = entry_peers or []

        self.mailbox_uri = mailbox_uri

        # peer configuration
        self.peer_registration_delay = peer_registration_delay
        self.records_storage_path = records_storage_path
        if (
            self.records_storage_path is not None
            and not Path(self.records_storage_path).is_absolute()
        ):
            self.records_storage_path = os.path.join(  # pragma: nocover
                data_dir, self.records_storage_path
            )

        # node startup
        self.source = os.path.abspath(module_path)
        self.clargs = clargs or []

        # node libp2p multiaddress
        self.multiaddrs: Sequence[MultiAddr] = []

        # log file
        self.log_file = log_file or LIBP2P_NODE_LOG_FILE
        if not Path(self.log_file).is_absolute():
            self.log_file = os.path.join(data_dir, self.log_file)  # pragma: nocover
        # env file
        self.env_file = env_file or LIBP2P_NODE_ENV_FILE
        if not Path(self.env_file).is_absolute():
            self.env_file = os.path.join(data_dir, self.env_file)

        # named pipes (fifos)
        self.pipe: Optional[IPCChannel] = None
        self._loop: Optional[AbstractEventLoop] = None
        self.proc: Optional[subprocess.Popen] = None
        self._log_file_desc: Optional[IO[str]] = None

        self.logger = logger
        self._connection_timeout = connection_timeout or PIPE_CONN_TIMEOUT
        self._max_restarts = max_restarts
        self._restart_counter: int = 0
        self._is_on_stop: bool = False

    def _make_env_file(self, pipe_in_path: str, pipe_out_path: str) -> str:
        # setup config
        if os.path.exists(self.env_file):
            os.remove(self.env_file)  # pragma: nocover

        config = ""
        config += "AEA_AGENT_ADDR={}\n".format(self.address)
        config += "AEA_P2P_ID={}\n".format(self.key)
        config += "AEA_P2P_URI={}\n".format(str(self.uri))
        config += "AEA_P2P_ENTRY_URIS={}\n".format(
            ",".join(
                [
                    str(maddr)
                    for maddr in self.entry_peers
                    if str(maddr) != str(self.uri)  # TOFIX(LR) won't exclude self
                ]
            )
        )
        config += "NODE_TO_AEA={}\n".format(pipe_in_path)
        config += "AEA_TO_NODE={}\n".format(pipe_out_path)
        config += "AEA_P2P_URI_PUBLIC={}\n".format(
            str(self.public_uri) if self.public_uri is not None else ""
        )
        config += "AEA_P2P_DELEGATE_URI={}\n".format(
            str(self.delegate_uri) if self.delegate_uri is not None else ""
        )
        config += "AEA_P2P_URI_MONITORING={}\n".format(
            str(self.monitoring_uri) if self.monitoring_uri is not None else ""
        )
        config += "AEA_P2P_POR_ADDRESS={}\n".format(self.record.address)
        config += "AEA_P2P_POR_PUBKEY={}\n".format(self.record.public_key)
        config += "AEA_P2P_POR_PEER_PUBKEY={}\n".format(
            self.record.representative_public_key
        )
        config += "AEA_P2P_POR_SIGNATURE={}\n".format(self.record.signature)
        config += "AEA_P2P_POR_SERVICE_ID={}\n".format(POR_DEFAULT_SERVICE_ID)
        config += "AEA_P2P_POR_LEDGER_ID={}\n".format(self.record.ledger_id)
        config += "AEA_P2P_CFG_REGISTRATION_DELAY={}\n".format(
            str(self.peer_registration_delay)
            if self.peer_registration_delay is not None
            else str(0.0)
        )
        config += "AEA_P2P_CFG_STORAGE_PATH={}\n".format(
            self.records_storage_path if self.records_storage_path is not None else ""
        )

        config += "AEA_P2P_MAILBOX_URI={}\n".format(self.mailbox_uri)

        with open(self.env_file, "w") as env_file:  # overwrite if exists
            env_file.write(config)

        return config

    async def _set_connection_to_node(self) -> bool:
        if self.pipe is None:
            raise RuntimeError("pipe was not set")  # pragma: nocover

        return await self.pipe.connect(timeout=self._connection_timeout)

    def get_client(self) -> NodeClient:
        """Get client instance to communicate to node."""
        if self.pipe is None:
            raise RuntimeError("pipe was not set")  # pragma: nocover

        return NodeClient(self.pipe, self.record)

    def _child_watcher_callback(self, *_) -> None:  # type: ignore # pragma: nocover
        """Log if process was terminated before stop was called."""
        if self._is_on_stop or self.proc is None:
            return
        self.proc.poll()
        returncode = self.proc.returncode
        self.logger.error(
            f"Node process with pid {self.proc.pid} was terminated with returncode {returncode}"
        )

    def is_proccess_running(self) -> bool:
        """Check process is running."""
        if not self.proc:
            return False

        self.proc.poll()
        return self.proc.returncode is None

    async def start(self) -> None:
        """Start the node."""
        self._is_on_stop = False
        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self._log_file_desc = open(self.log_file, "a", 1)

        # tcp socket on every platform
        self.pipe = TCPSocketChannel(logger=self.logger)

        env_file_data = self._make_env_file(
            pipe_in_path=self.pipe.in_path, pipe_out_path=self.pipe.out_path
        )
        # run node
        self.logger.info("Starting libp2p node...")
        self.proc = _golang_module_run(
            self.source, LIBP2P_NODE_MODULE_NAME, [self.env_file], self._log_file_desc
        )

        if (
            platform.system() != "Windows"
            and sys.version_info.major == 3
            and sys.version_info.minor >= 8
        ):  # pragma: nocover
            with events.get_child_watcher() as watcher:
                if watcher:
                    watcher.add_child_handler(
                        self.proc.pid, self._child_watcher_callback
                    )

        self.logger.info("Connecting to libp2p node...")

        try:
            connected = await self._set_connection_to_node()

            if not connected:
                raise TimeoutError("Couldn't connect to libp2p process within timeout")
        except Exception as e:
            err_msg = self.get_libp2p_node_error()
            self.logger.error("Couldn't connect to libp2p process: {}".format(err_msg))
            self.logger.error(
                "Libp2p process configuration:\n{}".format(env_file_data.strip())
            )
            if err_msg == "":
                with open(self.log_file, "r") as f:
                    self.logger.error(
                        "Libp2p process log file {}:\n{}".format(
                            self.log_file, f.read()
                        )
                    )
            else:  # pragma: nocover
                log_data = Path(self.log_file).read_text()
                self.logger.error(f"Failure to connect to Libp2pNode:\n{log_data}")

            await self.stop()
            raise e

        self.logger.info("Successfully connected to libp2p node!")
        self.multiaddrs = self.get_libp2p_node_multiaddrs()
        self.describe_configuration()

    async def restart(self) -> None:
        """Perform node restart."""
        if self._restart_counter >= self._max_restarts:
            raise ValueError(f"Max restarts attempts reached: {self._max_restarts}")
        await self.stop()
        await self.start()
        self._restart_counter += 1

    def describe_configuration(self) -> None:
        """Print a message describing the libp2p node configuration"""
        msg = LIBP2P_SUCCESS_MESSAGE

        if self.public_uri is not None:
            msg += "full DHT mode with "
            if self.delegate_uri is not None:  # pragma: nocover
                msg += "delegate service reachable at '{}:{}' and relay service enabled. ".format(
                    self.public_uri.host, self.delegate_uri.port
                )
            else:
                msg += "relay service enabled. "

            msg += "To join its network use multiaddr '{}'.".format(self.multiaddrs[0])
        else:
            msg += "relayed mode and cannot be used as entry peer."

        self.logger.info(msg)

    def get_libp2p_node_multiaddrs(self) -> Sequence[MultiAddr]:
        """
        Get the node's multiaddresses.

        :return: a list of multiaddresses
        """
        LIST_START = "MULTIADDRS_LIST_START"
        LIST_END = "MULTIADDRS_LIST_END"

        multiaddrs: List[MultiAddr] = []

        with open(self.log_file, "r") as f:
            lines = f.readlines()

        found = False
        for line in lines:
            if LIST_START in line:
                found = True
                multiaddrs = []
                continue
            if found:
                elem = line.strip()
                if elem != LIST_END and len(elem) != 0:
                    multiaddrs.append(MultiAddr.from_string(elem))
                else:
                    found = False
        return multiaddrs

    def get_libp2p_node_error(self) -> str:
        """
        Parses libp2p node logs for critical errors

        :return: error message if any, empty string otherwise
        """

        CRITICAL_ERROR = "LIBP2P_NODE_PANIC_ERROR"
        PANIC_ERROR = "panic:"

        error_msg = ""
        panic_msg = ""

        with open(self.log_file, "r") as f:
            lines = f.readlines()

        for line in lines:  # pragma: nocover
            if CRITICAL_ERROR in line:
                parts = line.split(":", 1)
                error_msg = parts[1].strip()
            if PANIC_ERROR in line:
                parts = line.split(":", 1)
                panic_msg = parts[1].strip()

        return error_msg if error_msg != "" else panic_msg

    async def stop(self) -> None:
        """Stop the node."""
        if self.proc is not None:
            self.logger.debug("Terminating node process {}...".format(self.proc.pid))
            self._is_on_stop = True
            self.proc.poll()
            self.proc.terminate()
            self.logger.debug(
                "Waiting for node process {} to terminate...".format(self.proc.pid)
            )
            self.proc.wait()
            if self._log_file_desc is None:
                raise ValueError("log file descriptor is not set.")  # pragma: nocover
            self._log_file_desc.close()
        else:
            self.logger.debug("Called stop when process not set!")  # pragma: no cover

        if self.pipe is not None:
            try:
                await self.pipe.close()
            except Exception as e:  # pragma: nocover pylint: disable=broad-except
                self.logger.exception(f"Failure during pipe closing. Exception: {e}")
            self.pipe = None
        else:
            self.logger.debug("Called stop when pipe not set!")  # pragma: no cover

        if os.path.exists(self.env_file):
            os.remove(self.env_file)


class P2PLibp2pConnection(Connection):
    """A libp2p p2p node connection."""

    connection_id = PUBLIC_ID
    DEFAULT_MAX_RESTARTS = 5

    def __init__(self, **kwargs: Any) -> None:
        """Initialize a p2p libp2p connection."""
        super().__init__(**kwargs)
        ledger_id = self.configuration.config.get("ledger_id", DEFAULT_LEDGER)
        if ledger_id not in SUPPORTED_LEDGER_IDS:
            raise ValueError(  # pragma: nocover
                "Ledger id '{}' is not supported. Supported ids: '{}'".format(
                    ledger_id, SUPPORTED_LEDGER_IDS
                )
            )
        libp2p_local_uri: Optional[str] = self.configuration.config.get("local_uri")
        libp2p_public_uri: Optional[str] = self.configuration.config.get("public_uri")
        libp2p_delegate_uri: Optional[str] = self.configuration.config.get(
            "delegate_uri"
        )
        libp2p_monitoring_uri: Optional[str] = self.configuration.config.get(
            "monitoring_uri"
        )
        libp2p_entry_peers = self.configuration.config.get("entry_peers")
        if libp2p_entry_peers is None:
            libp2p_entry_peers = []
        libp2p_entry_peers = list(cast(List, libp2p_entry_peers))
        log_file: Optional[str] = self.configuration.config.get("log_file")
        env_file: Optional[str] = self.configuration.config.get("env_file")
        peer_registration_delay: Optional[str] = self.configuration.config.get(
            "peer_registration_delay"
        )
        records_storage_path: Optional[str] = self.configuration.config.get(
            "storage_path"
        )
        node_connection_timeout: Optional[float] = self.configuration.config.get(
            "node_connection_timeout", PIPE_CONN_TIMEOUT
        )
        if (
            self.has_crypto_store
            and self.crypto_store.crypto_objects.get(ledger_id, None) is not None
        ):  # pragma: no cover
            key = self.crypto_store.crypto_objects[ledger_id]
        else:
            raise ValueError(
                f"Couldn't find connection key for {str(ledger_id)} in connections keys. "
                "Please ensure agent private key is added with `aea add-key`."
            )

        uri = None
        if libp2p_local_uri is not None:
            uri = Uri(libp2p_local_uri)

        public_uri = None
        if libp2p_public_uri is not None:
            public_uri = Uri(libp2p_public_uri)

        delegate_uri = None
        if libp2p_delegate_uri is not None:  # pragma: nocover
            delegate_uri = Uri(libp2p_delegate_uri)

        monitoring_uri = None
        if libp2p_monitoring_uri is not None:
            monitoring_uri = Uri(libp2p_monitoring_uri)  # pragma: nocover

        entry_peers = [
            MultiAddr.from_string(str(maddr)) for maddr in libp2p_entry_peers
        ]

        delay = None
        if peer_registration_delay is not None:
            try:
                delay = float(peer_registration_delay)
            except ValueError:
                raise ValueError(
                    f"peer_registration_delay {peer_registration_delay} must be a float number in seconds"
                )

        if public_uri is None:
            # node will be run as a ClientDHT
            # requires entry peers to use as relay
            if entry_peers is None or len(entry_peers) == 0:
                raise ValueError(  # pragma: no cover
                    "At least one Entry Peer should be provided when node is run in relayed mode"
                )
            if delegate_uri is not None:  # pragma: no cover
                self.logger.warning(
                    "Ignoring Delegate Uri configuration as node is run in relayed mode"
                )
        else:
            # node will be run as a full NodeDHT
            if uri is None:
                raise ValueError(  # pragma: no cover
                    "Local Uri must be set when Public Uri is provided. "
                    "Hint: they are the same for local host/network deployment"
                )
            # check if node's public host and entry peers hosts are either
            #  both private or both public
            if not _ip_all_private_or_all_public(
                [public_uri.host] + [maddr.host for maddr in entry_peers]
            ):
                raise ValueError(  # pragma: nocover
                    "Node's public ip and entry peers ip addresses are not in the same ip address space (private/public)"
                )

        cert_requests = self.configuration.cert_requests
        if cert_requests is None or len(cert_requests) != 1:
            raise ValueError(  # pragma: no cover
                "cert_requests field must be set and contain exactly one entry!"
            )
        cert_request = cert_requests[0]

        agent_record = AgentRecord.from_cert_request(
            cert_request, self.address, key.public_key, Path(self.data_dir)
        )

        # libp2p local node
        self.logger.debug("Public key used by libp2p node: {}".format(key.public_key))

        if self.configuration.config.get("mailbox_uri"):
            mailbox_uri = str(self.configuration.config.get("mailbox_uri"))
        else:
            mailbox_uri = ""

        module_dir = self._check_node_built()
        self.node = Libp2pNode(
            agent_record,
            key,
            module_dir,
            self.data_dir,
            LIBP2P_NODE_CLARGS,
            uri,
            public_uri,
            delegate_uri,
            monitoring_uri,
            entry_peers,
            log_file,
            env_file,
            self.logger,
            delay,
            records_storage_path,
            node_connection_timeout,
            max_restarts=self.configuration.config.get(
                "max_node_restarts", self.DEFAULT_MAX_RESTARTS
            ),
            mailbox_uri=mailbox_uri,
        )

        self._in_queue = None  # type: Optional[asyncio.Queue]
        self._receive_from_node_task = None  # type: Optional[asyncio.Future]
        self._node_client: Optional[NodeClient] = None

        self._send_queue: Optional[asyncio.Queue] = None
        self._send_task: Optional[asyncio.Task] = None

    def _check_node_built(self) -> str:
        """Check node built."""
        if self.configuration.build_directory is None:
            raise ValueError("Connection Configuration build directory is not set!")

        libp2p_node_module_path = os.path.join(
            self.configuration.build_directory, LIBP2P_NODE_MODULE_NAME
        )
        enforce(
            os.path.exists(libp2p_node_module_path),
            f"Module {LIBP2P_NODE_MODULE_NAME} is not present in {self.configuration.build_directory}, please call the `aea build` command first!",
        )
        return self.configuration.build_directory

    async def connect(self) -> None:
        """Set up the connection."""
        if self.is_connected:
            return  # pragma: nocover
            # start libp2p node
        with self._connect_context():
            self.node.logger = self.logger
            await self._start_node()
            # starting receiving msgs
            self._in_queue = asyncio.Queue()
            self._send_queue = asyncio.Queue()
            self._receive_from_node_task = asyncio.ensure_future(
                self._receive_from_node(), loop=self.loop
            )
            self._send_task = self.loop.create_task(self._send_loop())

    async def _start_node(self) -> None:
        """Start node and set node client instance."""
        await self.node.start()
        self._node_client = self.node.get_client()

    async def _restart_node(self) -> None:
        """Stop and start node again."""
        await self.node.stop()
        await self._start_node()

    async def disconnect(self) -> None:
        """Disconnect from the channel."""
        if self.is_disconnected:
            return  # pragma: nocover

        self.state = ConnectionStates.disconnecting
        try:
            if self._receive_from_node_task is not None:
                self._receive_from_node_task.cancel()
                self._receive_from_node_task = None

            if self._send_task is not None:
                self._send_task.cancel()
                self._send_task = None

            await self.node.stop()
            if self._in_queue is not None:
                self._in_queue.put_nowait(None)
            else:
                self.logger.debug(  # pragma: nocover
                    "Called disconnect when input queue not initialized."
                )
        finally:
            self.state = ConnectionStates.disconnected

    async def receive(self, *args: Any, **kwargs: Any) -> Optional["Envelope"]:
        """
        Receive an envelope. Blocking.

        :param args: positional arguments
        :param kwargs: keyword arguments
        :return: the envelope received, or None.
        """
        try:
            if self._in_queue is None:
                raise ValueError("Input queue not initialized.")  # pragma: nocover
            envelope = await self._in_queue.get()
            if envelope is None:  # pragma: nocover
                self.logger.debug("Received None.")
                return None
            return envelope
        except CancelledError:  # pragma: no cover
            self.logger.debug("Receive cancelled.")
            return None
        except Exception as e:  # pragma: nocover # pylint: disable=broad-except
            self.logger.exception(e)
            return None

    async def _send_envelope_with_node_client(self, envelope: Envelope) -> None:
        if not self._node_client:  # pragma: nocover
            raise ValueError(f"Node client not set! Can not send envelope: {envelope}")

        if not self.node.pipe:  # pragma: nocover
            raise ValueError("Node is not connected")

        try:
            await self._node_client.send_envelope(envelope)
            return
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise  # pragma: nocover
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception(
                f"Failed to send. Exception: {e}. Try recover connection to node and send again."
            )

        try:
            if self.node.is_proccess_running():
                await self.node.pipe.connect()
                await self._node_client.send_envelope(envelope)
                self.logger.debug("Envelope sent after reconnect to node")
                return
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise  # pragma: nocover
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception(
                f"Failed to send after pipe reconnect. Exception: {e}. Try recover connection to node and send again."
            )

        try:
            await self._restart_node()
            await self._node_client.send_envelope(envelope)
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise  # pragma: nocover
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception(
                f"Failed to send after node restart. Exception: {e}. Try recover connection to node and send again."
            )
            raise

    async def _send_loop(self) -> None:
        """Handle message in  the send queue."""

        if not self._send_queue or not self._node_client:  # pragma: nocover
            self.logger.error("Send loop not started cause not connected properly.")
            return
        try:
            while self.is_connected:
                envelope = await self._send_queue.get()
                await self._send_envelope_with_node_client(envelope)
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise  # pragma: nocover
        except Exception:  # pylint: disable=broad-except # pragma: nocover
            self.logger.exception(
                f"Failed to send an envelope {envelope}. Stop connection."
            )
            await asyncio.shield(self.disconnect())

    async def send(self, envelope: Envelope) -> None:
        """
        Send messages.

        :param envelope: the envelope
        """
        if not self._node_client or not self._send_queue:
            raise ValueError("Node is not connected!")  # pragma: nocover

        self._ensure_valid_envelope_for_external_comms(envelope)
        await self._send_queue.put(envelope)

    async def _read_envelope_from_node(self) -> Optional[Envelope]:
        if not self._node_client:
            raise ValueError("Node is not connected!")  # pragma: nocover
        try:
            return await self._node_client.read_envelope()
        except asyncio.CancelledError:  # pylint: disable=try-except-raise
            raise  # pragma: nocover
        except Exception as e:  # pylint: disable=broad-except
            self.logger.exception(
                f"Failed to read. Exception: {e}. Try reconnect to node and read again."
            )

            await self._restart_node()
            return await self._node_client.read_envelope()

    async def _receive_from_node(self) -> None:
        """Receive data from node."""
        while True:
            if self._in_queue is None:
                raise ValueError("Input queue not initialized.")  # pragma: nocover

            if not self._node_client:
                raise ValueError("Node is not connected!")  # pragma: nocover

            envelope = await self._read_envelope_from_node()
            if envelope is None:
                break
            self._in_queue.put_nowait(envelope)
